# coding=utf-8
"""
Field filters for WTForm.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from future.utils import string_types


def strip(data):
    """
    Strip data if data is a string
    """
    if data is None:
        return ""
    if not isinstance(data, string_types):
        return data
    return data.strip()


def uppercase(data):
    if not isinstance(data, string_types):
        return data
    return data.upper()


def lowercase(data):
    if not isinstance(data, string_types):
        return data
    return data.lower()



from __future__ import absolute_import

import tempfile
from os.path import dirname, join
from unittest import TestCase
from warnings import warn

from magic import Magic, os

from abilian.services.conversion import converter

BASEDIR = join(dirname(__file__), "..", "dummy_files")
BASEDIR2 = join(dirname(__file__), "..", "dummy_files2")

mime_sniffer = Magic(mime=True)
encoding_sniffer = Magic(mime_encoding=True)

# FIXME: tests that rely on OOo are disabled until we fix stability issues.


class Test(TestCase):

    @classmethod
    def setUpClass(cls):
        cache_dir = tempfile.mkdtemp(suffix='unittest')
        tmp_dir = tempfile.mkdtemp(suffix='unittest')
        converter.init_work_dirs(cache_dir, tmp_dir)

    @classmethod
    def tearDownClass(cls):
        converter.clear()

    def read_file(self, fn):
        try:
            return open(join(BASEDIR, fn)).read()
        except IOError as e:
            return open(join(BASEDIR2, fn)).read()

    # To text
    def test_pdf_to_text(self):
        if not os.popen("which pdftotex").read().strip():
            warn("pdftotext not found, skipping test")
            return
        blob = self.read_file("onepage.pdf")
        text = converter.to_text("", blob, "application/pdf")

    def XXXtest_word_to_text(self):
        blob = self.read_file("test.doc")
        text = converter.to_text("", blob, "application/msword")

    def XXXtest_wordx_to_text(self):
        blob = self.read_file("test.docx")
        text = converter.to_text("", blob, "application/msword")

    def XXXtest_excel_to_text(self):
        blob = self.read_file("test.xls")
        text = converter.to_text("", blob, "application/excel")

    # To PDF
    def XXXtest_odt_to_pdf(self):
        blob = self.read_file("test.odt")
        pdf = converter.to_pdf("", blob,
                               "application/vnd.oasis.opendocument.text")
        assert "application/pdf" == mime_sniffer.from_buffer(pdf)

    def XXXtest_word_to_pdf(self):
        blob = self.read_file("test.doc")
        pdf = converter.to_pdf("", blob, "application/msword")
        assert "application/pdf" == mime_sniffer.from_buffer(pdf)

    def test_image_to_pdf(self):
        blob = self.read_file("picture.jpg")
        pdf = converter.to_pdf("", blob, "image/jpeg")
        assert "application/pdf" == mime_sniffer.from_buffer(pdf)

    # To images
    def test_pdf_to_images(self):
        if not os.popen("which pdftoppm").read().strip():
            warn("pdftoppm not found, skipping test")
            return
        blob = self.read_file("onepage.pdf")
        image = converter.to_image("", blob, "application/pdf", 0)
        assert "image/jpeg" == mime_sniffer.from_buffer(image)

    def XXXtest_word_to_images(self):
        blob = self.read_file("test.doc")
        image = converter.to_image("", blob, "application/msword", 0)
        assert "image/jpeg" == mime_sniffer.from_buffer(image)



from __future__ import absolute_import, print_function, unicode_literals

from os.path import dirname, join

from pytest import fixture

from ..image import CROP, FIT, SCALE, get_save_format, get_size, resize


@fixture
def orig_image():
    # 725x518
    return open(join(dirname(__file__), "cat.jpg")).read()


def test_get_save_format():
    assert get_save_format('JPG') == 'JPEG'
    assert get_save_format('JPEG') == 'JPEG'
    assert get_save_format('PNG') == 'PNG'
    assert get_save_format('GIF') == 'PNG'
    assert get_save_format('unknown') == 'JPEG'


def test_fit(orig_image):
    image = resize(orig_image, 500, 500, FIT)
    x, y = get_size(image)
    assert (x, y) == (500, 357)

    # image already fits in desired dimension
    image = resize(orig_image, 1000, 1000, FIT)
    x, y = get_size(image)
    assert x == 725 and y == 518


def test_scale(orig_image):
    image = resize(orig_image, 500, 500, SCALE)
    assert get_size(image) == (500, 500)


def test_crop(orig_image):
    image = resize(orig_image, 500, 500, CROP)
    assert get_size(image) == (500, 500)



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from datetime import datetime
from unittest import TestCase

import sqlalchemy as sa

from abilian.core.entities import Entity
from abilian.core.models.base import (AUDITABLE, NOT_SEARCHABLE, SEARCHABLE,
                                      Info)
from abilian.core.models.subjects import User
from abilian.services import security
from abilian.testing import BaseTestCase as AbilianTestCase

from .dummy import DummyContact


class EntityTestCase(TestCase):

    def get_session(self):
        engine = sa.create_engine('sqlite:///:memory:', echo=False)
        Session = sa.orm.sessionmaker(bind=engine)
        session = Session()

        # flask-sqlalchemy as listeners looking for this
        session._model_changes = {}

        DummyContact.metadata.create_all(engine)
        return session

    def test(self):
        contact = DummyContact(first_name=u"John")
        assert contact.creator is None
        assert contact.owner is None

        user = User()
        contact.owner = user
        contact.creator = user

    def test_auto_slug_property(self):
        session = self.get_session()
        obj = DummyContact(name=u'a b c')
        session.add(obj)
        session.flush()
        assert obj.auto_slug == u'a-b-c'

        obj.name = u"C'est l'été !"
        assert obj.auto_slug == u'c-est-l-ete'

        # with a special space character
        obj.name = u"a_b\u205fc"  # U+205F: MEDIUM MATHEMATICAL SPACE
        assert obj.auto_slug == u'a-b-c'

        # with non-ascii translatable chars, like EN DASH U+2013 (–) and EM DASH
        # U+2014 (—). Standard separator is \u002d (\x2d) "-" HYPHEN-MINUS.
        # this test may fails depending on how  unicode normalization + char
        # substitution is done (order matters).
        obj.name = u'a\u2013b\u2014c'  # u'a–b—c'
        slug = obj.auto_slug
        assert slug == u'a-b-c'
        assert u'\u2013' not in slug
        assert u'\u002d' in slug

    def test_updated_at(self):
        session = self.get_session()
        contact = DummyContact()
        session.add(contact)
        session.commit()
        assert isinstance(contact.updated_at, datetime)

        updated = contact.updated_at
        contact.first_name = u'John'
        session.commit()
        assert isinstance(contact.updated_at, datetime)
        assert contact.updated_at > updated

    def test_auto_slug(self):
        session = self.get_session()
        contact1 = DummyContact(name=u'Pacôme Hégésippe Adélard Ladislas')
        session.add(contact1)
        session.flush()
        assert contact1.slug == u'pacome-hegesippe-adelard-ladislas'

        # test when name is None
        contact2 = DummyContact()
        session.add(contact2)
        session.flush()
        expected = u'dummycontact-{}'.format(contact2.id)
        assert contact2.slug == expected

        # test numbering if slug already exists:
        contact3 = DummyContact(name=u'Pacôme Hégésippe Adélard Ladislas')
        session.add(contact3)
        session.flush()
        assert contact3.slug == u'pacome-hegesippe-adelard-ladislas-1'

    def test_polymorphic_update_timestamp(self):
        session = self.get_session()
        contact = DummyContact(name=u'Pacôme Hégésippe Adélard Ladislas')
        session.add(contact)
        session.flush()

        updated_at = contact.updated_at
        assert updated_at
        contact.email = u'p@example.com'
        session.flush()
        assert contact.updated_at > updated_at

    def test_meta(self):
        session = self.get_session()
        e = DummyContact(name=u'test')
        e.meta['key'] = u'value'
        e.meta['number'] = 42
        session.add(e)
        session.flush()
        e_id = e.id
        session.expunge(e)
        del e
        e = session.query(DummyContact).get(e_id)
        assert e.meta['key'] == u'value'
        assert e.meta['number'] == 42

    def test_entity_type(self):

        class MyType(Entity):
            pass

        expected = __name__ + '.MyType'
        self.assertEquals(MyType.entity_type, expected)
        self.assertEquals(MyType._object_type(), expected)

        class Fixed(Entity):
            entity_type = 'some.fixed.module.fixed_type'

        assert Fixed.entity_type == 'some.fixed.module.fixed_type'
        assert Fixed._object_type() == 'some.fixed.module.fixed_type'

        class OtherBase(Entity):
            ENTITY_TYPE_BASE = 'some.module'

        assert OtherBase.entity_type == 'some.module.OtherBase'
        assert OtherBase._object_type() == 'some.module.OtherBase'

        # test when ENTITY_TYPE_BASE is in ancestors
        class Base(object):
            ENTITY_TYPE_BASE = 'from.ancestor'

        class InheritedBase(Base, Entity):
            pass

        assert InheritedBase.entity_type == 'from.ancestor.InheritedBase'
        assert InheritedBase._object_type() == 'from.ancestor.InheritedBase'


class PermissionsTestCase(AbilianTestCase):

    def test_default_permissions(self):

        class MyRestrictedType(Entity):
            __default_permissions__ = {
                security.READ: {security.Anonymous},
                security.WRITE: {security.Owner},
                security.CREATE: {security.Writer},
                security.DELETE: {security.Owner},
            }

        assert isinstance(MyRestrictedType.__default_permissions__, frozenset)
        expected = frozenset(((security.READ, frozenset((security.Anonymous,))),
                              (security.WRITE, frozenset((security.Owner,))),
                              (security.CREATE, frozenset((security.Writer,))),
                              (security.DELETE, frozenset((security.Owner,))),))
        assert MyRestrictedType.__default_permissions__ == expected

        self.app.db.create_all()  # create missing 'mytype' table

        obj = MyRestrictedType(name=u'test object')
        self.session.add(obj)
        PA = security.PermissionAssignment
        query = self.session.query(PA.role) \
            .filter(PA.object == obj)

        assert query.filter(PA.permission == security.READ).all() \
               == [(security.Anonymous,)]

        assert query.filter(PA.permission == security.WRITE).all() \
               == [(security.Owner,)]

        assert query.filter(PA.permission == security.DELETE).all() \
               == [(security.Owner,)]

        # special case:
        assert query.filter(PA.permission == security.CREATE).all() \
               == []

        svc = self.app.services['security']
        permissions = svc.get_permissions_assignments(obj)
        assert permissions == {
            security.READ: {security.Anonymous},
            security.WRITE: {security.Owner},
            security.DELETE: {security.Owner},
        }


def test_info():
    info = SEARCHABLE
    assert info['searchable']

    info = NOT_SEARCHABLE
    assert not info['searchable']

    info = SEARCHABLE + AUDITABLE
    assert info['searchable']
    assert info['auditable']
    assert isinstance(info, Info)

    info = SEARCHABLE | AUDITABLE
    assert info['searchable']
    assert info['auditable']
    assert isinstance(info, Info)



# coding=utf-8

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from functools import wraps

from flask import Blueprint, current_app, jsonify, request
from flask_wtf import Form
from werkzeug.exceptions import Forbidden

blueprint = Blueprint('csrf', __name__, url_prefix='/csrf')


@blueprint.route('/token', endpoint='json_token')
def json_token_view():
    return jsonify(token=token())


def field():
    """
    Return an instance of `wtforms.ext.csrf.fields.CSRFTokenField`, suitable for
    rendering. Renders an empty string if `config.CSRF_ENABLED` is not set.
    """
    return Form().csrf_token


def time_limit():
    """Return current time limit for CSRF token.
    """
    return current_app.config.get('WTF_CSRF_TIME_LIMIT', 3600)


def name():
    """
    Field name expected to have CSRF token. Useful for passing it to
    JavaScript for instance.
    """
    return u'csrf_token'


def token():
    """
    Value of current csrf token. Useful for passing it to JavaScript for
    instance.
    """
    return field().current_token or u''


def support_graceful_failure(view):
    """Decorator to indicate that the view will handle itself the csrf failure.

    View can be a view function or a class based view
    """
    view.csrf_support_graceful_failure = True
    return view


def has_failed():
    return getattr(request, 'csrf_failed', False)


def protect(view):
    """Protects a view agains CSRF attacks by checking `csrf_token` value in
    submitted values. Do nothing if `config.CSRF_ENABLED` is not set.

    Raises `werkzeug.exceptions.Forbidden` if validation fails.
    """

    @wraps(view)
    def csrf_check(*args, **kwargs):
        # an empty form is used to validate current csrf token and only that!
        if not Form().validate():
            raise Forbidden('CSRF validation failed.')

        return view(*args, **kwargs)

    return csrf_check



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from datetime import datetime

from sqlalchemy.schema import Column
from sqlalchemy.types import DateTime, Integer
from whoosh.fields import ID

from ..extensions import db
from ..util import fqcn

#: Base Model class.
Model = db.Model


class Info(dict):

    def __init__(self, **kw):
        for k, v in kw.items():
            self[k] = v

    def copy(self):
        # dict.copy would return an instance of dict
        return self.__class__(**self)

    def __add__(self, other):
        d = self.copy()
        d.update(other)
        return d

    __or__ = __add__


EDITABLE = Info(editable=True)
NOT_EDITABLE = Info(editable=False)
AUDITABLE = Info(auditable=True)
AUDITABLE_HIDDEN = Info(auditable=True, audit_hide_content=True)
NOT_AUDITABLE = Info(auditable=False)
SEARCHABLE = Info(searchable=True)
NOT_SEARCHABLE = Info(searchable=False)
EXPORTABLE = Info(exportable=True)
NOT_EXPORTABLE = Info(exportable=False)

#: SYSTEM properties are properties defined by the system
#: and not supposed to be changed manually.
SYSTEM = Info(editable=False, auditable=False)


class IdMixin(object):
    id = Column(Integer, primary_key=True, info=SYSTEM | SEARCHABLE)


class Indexable(object):
    """
    Mixin with sensible defaults for indexable objects.
    """
    __indexable__ = True
    __indexation_args__ = {
        'index_to': (('object_key', (('object_key', ID(stored=True,
                                                       unique=True)),)),
                     ('object_type', (('object_type', ID(stored=True,
                                                         unique=False)),)),),
    }

    @classmethod
    def _object_type(cls):
        return fqcn(cls)

    @property
    def object_type(self):
        return self._object_type()

    @property
    def object_key(self):
        return u'{}:{}'.format(self.object_type, self.id)


class TimestampedMixin(object):
    #: creation date
    created_at = Column(DateTime,
                        default=datetime.utcnow,
                        info=SYSTEM | SEARCHABLE)
    #: last modification date
    updated_at = Column(DateTime,
                        default=datetime.utcnow,
                        onupdate=datetime.utcnow,
                        info=SYSTEM | SEARCHABLE)
    deleted_at = Column(DateTime, default=None, info=SYSTEM)



"""
A few utility functions.

See https://docs.djangoproject.com/en/dev/topics/http/shortcuts/ for more ideas
of stuff to implement.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import sys

from flask import url_for as flask_url_for
from flask import current_app, session, stream_with_context
from flask.helpers import send_from_directory
from future.utils import string_types
from werkzeug.routing import BuildError

try:
    import ipdb as pdb
except ImportError:
    import pdb


def url_for(obj, **kw):
    """Polymorphic variant of Flask's `url_for` function.

    Behaves like the original function when the first argument is a string.
    When it's an object, it
    """
    if isinstance(obj, string_types):
        return flask_url_for(obj, **kw)

    try:
        return current_app.default_view.url_for(obj, **kw)
    except KeyError:
        if hasattr(obj, "_url"):
            return obj._url
        elif hasattr(obj, "url"):
            return obj.url

    raise BuildError(repr(obj), kw, 'GET')


def get_object_or_404(cls, *args):
    """Shorthand similar to Django's `get_object_or_404`.
    """

    return cls.query.filter(*args).first_or_404()


def send_file_from_directory(filename, directory, app=None):
    """Helper to add static rules, like in `abilian.app`.app

    Example use::

       app.add_url_rule(
          app.static_url_path + '/abilian/<path:filename>',
          endpoint='abilian_static',
          view_func=partial(send_file_from_directory,
                            directory='/path/to/static/files/dir'))
    """
    if app is None:
        app = current_app
    cache_timeout = app.get_send_file_max_age(filename)
    return send_from_directory(directory, filename, cache_timeout=cache_timeout)


def capture_stream_errors(logger, msg):
    """Decorator that capture and log errors during streamed response.

    Decorated function is automatically decorated with
    :func:<`Flask.stream_with_context`>.

    @param logger: a logger name or logger instance
    @param msg: message to log
    """
    if isinstance(logger, string_types):
        logger = logging.getLogger(logger)

    def decorator(fun):

        @stream_with_context
        def wrapper(*args, **kwargs):
            # this is for developpers convenience. The debugger middleware doesn't
            # work when using streamed responses.
            should_pdb = current_app.debug and session.get(
                'pdb_streamed_responses')
            try:
                generator = fun(*args, **kwargs)
                for chunk in generator:
                    yield chunk
            except Exception:
                # Anonymous "except" would also capture GeneratorExit that is not a
                # subclass of Exception. GeneratorExit happens when user stop download
                # during generation: it's not an error that should be logged
                type_, value, tb = sys.exc_info()
                if tb.tb_next is not None:
                    # error has happened inside decorated function, remove us from top
                    # stack: better readability in logs, accurate label in sentry
                    tb = tb.tb_next

                logger.error(msg, exc_info=(type_, value, tb))
                if should_pdb:
                    pdb.post_mortem(tb)

                raise

        return wrapper

    return decorator



# coding=utf-8
"""
Login-related views (login / logout / password reminder / ...).

Notes:
- Uses code copy/pasted (and modified) from Flask-Security
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import random
import string
from datetime import datetime
from urlparse import urljoin, urlparse

from flask import (current_app, flash, redirect, render_template, request,
                   url_for)
from flask.globals import _request_ctx_stack
from flask.json import jsonify
from flask_login import (login_user, logout_user, user_logged_in,
                         user_logged_out)
from flask_mail import Message
from flask_security.utils import md5
from itsdangerous import BadSignature, SignatureExpired, URLSafeTimedSerializer
from sqlalchemy import sql
from sqlalchemy.orm.exc import NoResultFound

from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.i18n import _, render_template_i18n
from abilian.services.security import Anonymous
from abilian.web.blueprints import Blueprint

from .models import LoginSession

__all__ = []

login = Blueprint("login",
                  __name__,
                  url_prefix="/user",
                  allowed_roles=Anonymous,
                  template_folder='templates')
route = login.route


#
# Login / Logout
#
@route("/login")
def login_form():
    """Display the login form."""
    next_url = get_redirect_target()
    return render_template("login/login.html", next_url=next_url)


def do_login(form):
    email = form.get('email', "").lower()
    password = form.get('password')
    next_url = form.get('next', u'')
    res = dict(username=email, email=email, next_url=next_url)

    if not email or not password:
        res['error'] = _(u"You must provide your email and password.")
        res['code'] = 401
        return res

    try:
        user = User.query \
            .filter(sql.func.lower(User.email) == email,
                    User.can_login == True) \
            .one()
    except NoResultFound:
        res['error'] = _(u"Sorry, we couldn't find an account for "
                         u"email '{email}'.").format(email=email)
        res['code'] = 401
        return res

    if user and not user.authenticate(password):
        res['error'] = _(u"Sorry, wrong password.")
        res['code'] = 401
        return res

    # Login successful
    login_user(user)
    res['user'] = user
    res['email'] = user.email
    return res


@csrf.exempt
@route("/login", methods=['POST'])
def login_post():
    res = do_login(request.form)
    if 'error' in res:
        code = res.pop('code')
        flash(res['error'], 'error')
        return (render_template("login/login.html", **res), code)

    return redirect_back(url=request.url_root)


@csrf.exempt
@route("/api/login", methods=['POST'])
def login_json():
    res = do_login(request.json)
    code = None

    if 'error' in res:
        code = res.pop('code')
    else:
        user = res.pop('user')
        res['fullname'] = user.name

    response = jsonify(**res)
    if code:
        response.status_code = code
    return response


@csrf.exempt
@route("/logout", methods=['GET', 'POST'])
def logout():
    logout_user()
    return redirect(request.url_root)


@csrf.exempt
@route("/api/logout", methods=['POST'])
def logout_json():
    logout_user()
    return u'Logged out, ok.', 200


#
# Code to deal with forgotten passwords or initial password generation
# Note: most of this code is stolen (copy/pasted) from Flask-Security, with a
# few modifications.
#
@route("/forgotten_pw")
def forgotten_pw_form():
    return render_template("login/forgotten_password.html")


@route("/forgotten_pw", methods=['POST'])
@csrf.exempt
def forgotten_pw(new_user=False):
    """
    Reset password for users who have already activated their accounts.
    """
    email = request.form.get('email', "").lower()

    action = request.form.get('action')
    if action == "cancel":
        return redirect(url_for("login.login_form"))

    if not email:
        flash(_(u"You must provide your email address."), 'error')
        return render_template("login/forgotten_password.html")

    try:
        user = User.query \
            .filter(sql.func.lower(User.email) == email,
                    User.can_login == True) \
            .one()
    except NoResultFound:
        flash(_(u"Sorry, we couldn't find an account for "
                "email '{email}'.").format(email=email), 'error')
        return render_template("login/forgotten_password.html"), 401

    if user.can_login and not user.password:
        user.set_password(random_password())
        db.session.commit()

    send_reset_password_instructions(user)
    flash(_("Password reset instructions have been sent to your email address."),
          'info')

    return redirect(url_for("login.login_form"))


@route("/reset_password/<token>")
def reset_password(token):
    expired, invalid, user = reset_password_token_status(token)
    if invalid:
        flash(_(u"Invalid reset password token."), "error")
    elif expired:
        flash(_(u"Password reset expired"), "error")
    if invalid or expired:
        return redirect(url_for('login.forgotten_pw'))

    return render_template_i18n("login/password_reset.html")


@route("/reset_password/<token>", methods=['POST'])
@csrf.exempt
def reset_password_post(token):
    action = request.form.get('action')
    if action == "cancel":
        return redirect(request.url_root)

    expired, invalid, user = reset_password_token_status(token)

    if invalid or not user:
        flash(_(u"Invalid reset password token."), "error")
    elif expired:
        flash(_(u"Password reset expired"), "error")
    if invalid or expired or not user:
        return redirect(url_for('login.forgotten_pw'))

    password = request.form.get('password')

    if not password:
        flash(_(u"You must provide a password."), "error")
        return redirect(url_for("login.reset_password_post", token=token))

    # TODO: check entropy
    if len(password) < 7:
        flash(_(u"Your new password must be at least 8 characters long"), "error")
        return redirect(url_for("login.reset_password_post", token=token))

    if password.lower() == password:
        flash(_(u"Your new password must contain upper case and lower case "
                "letters"), "error")
        return redirect(url_for("login.reset_password_post", token=token))

    if not len([x for x in password if x.isdigit()]) > 0:
        flash(_(u"Your new password must contain at least one digit"), "error")
        return redirect(url_for("login.reset_password_post", token=token))

    user.set_password(password)
    db.session.commit()

    flash(_(u"Your password has been changed. "
            u"You can now login with your new password"),
          "success")

    return redirect(url_for("login.login_form", next_url=request.url_root))


def random_password():
    pw = []
    for i in range(0, 10):
        pw.append(random.choice(string.letters + string.digits))
    return pw


def get_serializer(name):
    """
    :type name: str
    """
    config = current_app.config
    secret_key = config.get('SECRET_KEY')
    salt = config.get('SECURITY_{}_SALT'.format(name.upper()))
    return URLSafeTimedSerializer(secret_key=secret_key, salt=salt)


def send_reset_password_instructions(user):
    """Send the reset password instructions email for the specified user.

    :param user: The user to send the instructions to
    :type user: User
    """
    token = generate_reset_password_token(user)
    url = url_for('login.reset_password', token=token)
    reset_link = request.url_root[:-1] + url

    subject = _(u"Password reset instruction for {site_name}"
                ).format(site_name=current_app.config.get('SITE_NAME'))
    mail_template = 'password_reset_instructions'
    send_mail(subject,
              user.email,
              mail_template,
              user=user,
              reset_link=reset_link)


def generate_reset_password_token(user):
    """Generate a unique reset password token for the specified user.

    :param user: The user to work with
    :type user: User
    """
    data = [str(user.id), md5(user.password)]
    return get_serializer("reset").dumps(data)


def reset_password_token_status(token):
    """Return the expired status, invalid status, and user of a password reset
  token.

    For example::

        expired, invalid, user = reset_password_token_status('...')

    :param token: The password reset token
    :type token: str
    """
    return get_token_status(token, 'reset', 'RESET_PASSWORD')


def get_token_status(token, serializer_name, max_age=None):
    serializer = get_serializer(serializer_name)
    #max_age = get_max_age(max_age)
    user, data = None, None
    expired, invalid = False, False

    try:
        data = serializer.loads(token, max_age=max_age)
    except SignatureExpired:
        d, data = serializer.loads_unsafe(token)
        expired = True
    except BadSignature:
        invalid = True

    if data:
        user = User.query.get(data[0])

    expired = expired and (user is not None)
    return expired, invalid, user


def send_mail(subject, recipient, template, **context):
    """Send an email using the Flask-Mail extension.

    :param subject: Email subject
    :param recipient: Email recipient
    :param template: The name of the email template
    :param context: The context to render the template with
    """

    config = current_app.config
    sender = config['MAIL_SENDER']
    msg = Message(subject, sender=sender, recipients=[recipient])

    ctx = ('login/email', template)
    msg.body = render_template_i18n('%s/%s.txt' % ctx, **context)
    #msg.html = render_template('%s/%s.html' % ctx, **context)

    mail = current_app.extensions.get('mail')
    current_app.logger.debug("Sending mail...")
    mail.send(msg)


#
# Logging
#
def log_session_start(app, user):
    session = LoginSession.new()
    db.session.add(session)
    db.session.commit()


def log_session_end(app, user):
    session = LoginSession.query.get_active_for(user)
    if session:
        session.ended_at = datetime.utcnow()
        db.session.commit()


user_logged_in.connect(log_session_start)
user_logged_out.connect(log_session_end)


# login redirect utilities
#  from http://flask.pocoo.org/snippets/62/
def is_safe_url(target):
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.url_root, target))
    return test_url.scheme in ('http', 'https') and \
        ref_url.netloc == test_url.netloc


def check_for_redirect(target):
    target = urljoin(request.url_root, target)
    url = urlparse(target)
    reqctx = _request_ctx_stack.top
    try:
        endpoint, ignored = reqctx.url_adapter.match(url.path, 'GET')
        if '.' in endpoint and endpoint.rsplit('.', 1)[0] == 'login':
            # don't redirect to any login view after successful login
            return None
    except Exception as e:
        # exceptions may happen if route is not found for example
        pass

    return target


def get_redirect_target():
    for target in (request.values.get('next'), request.referrer):
        if not target:
            continue
        if is_safe_url(target):
            return check_for_redirect(target)


def redirect_back(endpoint=None, url=None, **values):
    target = request.form.get('next')
    if not target or not is_safe_url(target):
        if endpoint:
            target = url_for(endpoint, **values)
        else:
            target = url
    return redirect(check_for_redirect(target))



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import pygeoip
from flask import render_template

from abilian.i18n import _
from abilian.services.auth.models import LoginSession

from ..panel import AdminPanel

DATA_FILES = ('/usr/share/GeoIP/GeoIP.dat', '/usr/share/GeoIP/GeoIPv6.dat',)


class LoginSessionsPanel(AdminPanel):
    id = "login_sessions"
    label = "Session log"
    icon = 'log-in'

    def get(self):
        geoips = []
        for filename in DATA_FILES:
            try:
                geoips.append(pygeoip.GeoIP(filename))
            except (pygeoip.GeoIPError, IOError):
                pass

        sessions = LoginSession.query.order_by(LoginSession.id.desc()).limit(
            50).all()
        unknown_country = _(u'Country unknown')

        for session in sessions:
            country = unknown_country

            if geoips and session.ip_address:
                ip_address = session.ip_address
                multiple = ip_address.split(',')
                if multiple:
                    # only use last ip in the list, most likely the public address
                    ip_address = multiple[-1]
                for g in geoips:
                    try:
                        country = g.country_name_by_addr(ip_address)
                    except:  # noqa
                        continue

                    if country:
                        break
                    else:
                        country = unknown_country

                session.country = country

        return render_template("admin/login_sessions.html", **locals())



# coding=utf-8
"""
Extension for sending informations to upstream server
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from flask import _request_ctx_stack
from flask.signals import request_finished, request_started

from abilian.core.signals import user_loaded


class UpstreamInfo(object):
    """Extension for sending informations to upstream server
    """

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.extensions['upstream'] = self
        request_started.connect(self.request_started, sender=app)
        request_finished.connect(self.request_finished, sender=app)
        user_loaded.connect(self.user_loaded, sender=app)

        config = app.config
        config.setdefault('ABILIAN_UPSTREAM_INFO_ENABLED', False)
        for key in (u'ABILIAN_UPSTREAM_INFO_DISCARD',
                    u'ABILIAN_UPSTREAM_INFO_INCLUDE'):
            val = config.get(key, ())
            if val is not None:
                val = frozenset(val)
            config[key] = val

    def request_started(self, app):
        _request_ctx_stack.top.upstream_info = {'Username': u'Anonymous',}

    def request_finished(self, app, response):
        info = _request_ctx_stack.top.upstream_info
        config = app.config
        default_enabled = bool(config['ABILIAN_UPSTREAM_INFO_ENABLED'])
        disabled = config['ABILIAN_UPSTREAM_INFO_DISCARD']
        enabled = config['ABILIAN_UPSTREAM_INFO_INCLUDE']

        for key, val in info.items():
            if (default_enabled and key in disabled or not default_enabled and
                    key not in enabled):
                continue

            header = u'X-' + key
            response.headers[header] = val

    def user_loaded(self, app, user, *args, **kwargs):
        _request_ctx_stack.top.upstream_info['Username'] = user.email


extension = UpstreamInfo()



# coding=utf-8
"""
Base Flask application class, used by tests or to be extended
in real applications.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import errno
import importlib
import logging
import logging.config
import os
from functools import partial
from itertools import chain, count

import jinja2
import sqlalchemy as sa
import yaml
from babel.dates import LOCALTZ
from flask import (Blueprint, Flask, _request_ctx_stack, abort,
                   appcontext_pushed, current_app, g, render_template, request,
                   request_started)
from flask.config import ConfigAttribute
from flask.helpers import locked_cached_property
from flask_assets import Environment as AssetsEnv
from flask_assets import Bundle
from flask_babel import get_locale as babel_get_locale
from flask_migrate import Migrate
from flask_script import Manager as ScriptManager
from future.utils import string_types
from pathlib import Path
from pkg_resources import resource_filename
from sqlalchemy.orm.attributes import NEVER_SET, NO_VALUE
from werkzeug.datastructures import ImmutableDict
from werkzeug.utils import import_string

import abilian.core.util
import abilian.i18n
from abilian.core import extensions, redis, signals
from abilian.core.celery import FlaskCelery
from abilian.plugin.loader import AppLoader
from abilian.services import converter as conversion_service
from abilian.services import (
    activity_service, antivirus, audit_service, auth_service, index_service,
    preferences_service, repository_service, security_service,
    session_repository_service, settings_service, vocabularies_service)
from abilian.services.security import Anonymous
from abilian.web import csrf
from abilian.web.action import Endpoint, actions
from abilian.web.admin import Admin
from abilian.web.assets.filters import ClosureJS
from abilian.web.blueprints import allow_access_for_roles
from abilian.web.filters import init_filters
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import send_file_from_directory, url_for
from abilian.web.views import Registry as ViewRegistry
from abilian.web.views.images import user_photo_url

logger = logging.getLogger(__name__)
db = extensions.db
__all__ = ['create_app', 'Application', 'ServiceManager']


class ServiceManager(object):
    """
    Mixin that provides lifecycle (register/start/stop) support for services.
    """

    def __init__(self):
        self.services = {}

    def start_services(self):
        for svc in self.services.values():
            svc.start()

    def stop_services(self):
        for svc in self.services.values():
            svc.stop()


class PluginManager(object):
    """
    Mixin that provides support for loading plugins.
    """

    def load_plugins(self):
        """Discover and load plugins.

        At this point, prefer explicit loading using the :method:~`register_plugin`
        method.
        """
        loader = AppLoader()
        loader.load(__name__.split('.')[0])
        loader.register(self)

    def register_plugin(self, name):
        """Load and register a plugin given its package name.
        """
        logger.info("Registering plugin: " + name)
        module = importlib.import_module(name)
        module.register_plugin(self)


default_config = dict(Flask.default_config)
default_config.update(
    PRIVATE_SITE=False,
    TEMPLATE_DEBUG=False,
    CSRF_ENABLED=True,
    BABEL_ACCEPT_LANGUAGES=None,
    DEFAULT_COUNTRY=None,
    PLUGINS=(),
    ADMIN_PANELS=(
        'abilian.web.admin.panels.dashboard.DashboardPanel',
        'abilian.web.admin.panels.audit.AuditPanel',
        'abilian.web.admin.panels.login_sessions.LoginSessionsPanel',
        'abilian.web.admin.panels.settings.SettingsPanel',
        'abilian.web.admin.panels.users.UsersPanel',
        'abilian.web.admin.panels.groups.GroupsPanel',
        'abilian.web.admin.panels.sysinfo.SysinfoPanel',
        'abilian.services.vocabularies.admin.VocabularyPanel',
        'abilian.web.tags.admin.TagPanel',
    ),
    CELERYD_MAX_TASKS_PER_CHILD=1000,
    CELERY_ACCEPT_CONTENT=['pickle', 'json', 'msgpack', 'yaml'],
    CELERY_TIMEZONE=LOCALTZ,
    SENTRY_USER_ATTRS=('email', 'first_name', 'last_name',),
    SENTRY_INSTALL_CLIENT_JS=True,  # also install client JS
    SENTRY_JS_VERSION='1.1.22',
    SENTRY_JS_PLUGINS=('console', 'jquery', 'native', 'require',),
    SESSION_COOKIE_NAME=None,
    SQLALCHEMY_POOL_RECYCLE=1800,  # 30min. default value in flask_sa is None
    SQLALCHEMY_TRACK_MODIFICATIONS=False,
    LOGO_URL=Endpoint('abilian_static', filename='img/logo-abilian-32x32.png'),
    ABILIAN_UPSTREAM_INFO_ENABLED=False,  # upstream info extension
    TRACKING_CODE_SNIPPET=u'',  # tracking code to insert before </body>
    MAIL_ADDRESS_TAG_CHAR=None,
)
default_config = ImmutableDict(default_config)


class Application(Flask, ServiceManager, PluginManager):
    """Base application class. Extend it in your own app.
    """
    default_config = default_config

    #: Custom apps may want to always load some plugins: list them here.
    APP_PLUGINS = ('abilian.web.search',
                   'abilian.web.tags',
                   'abilian.web.comments',
                   'abilian.web.uploads',
                   'abilian.web.attachments',)

    #: Environment variable used to locate a config file to load last (after
    #: instance config file). Use this if you want to override some settings on a
    #: configured instance.
    CONFIG_ENVVAR = 'ABILIAN_CONFIG'

    #: True if application has a config file and can be considered configured for
    #: site.
    configured = ConfigAttribute('CONFIGURED')

    #: If True all views will require by default an authenticated user, unless
    #: Anonymous role is authorized. Static assets are always public.
    private_site = ConfigAttribute('PRIVATE_SITE')

    #: instance of :class:`.web.views.registry.Registry`.
    default_view = None

    #: json serializable dict to land in Javascript under Abilian.api
    js_api = None

    #: :class:`flask.ext.script.Manager` instance for shell commands of this app.
    #: defaults to `.commands.manager`, relative to app name.
    script_manager = '.commands.manager'

    #: celery app class
    celery_app_cls = FlaskCelery

    def __init__(self, name=None, config=None, *args, **kwargs):
        kwargs.setdefault('instance_relative_config', True)
        name = name or __name__

        # used by make_config to determine if we try to load config from instance /
        # environment variable /...
        self._ABILIAN_INIT_TESTING_FLAG = (getattr(config, 'TESTING', False) if
                                           config else False)
        Flask.__init__(self, name, *args, **kwargs)
        del self._ABILIAN_INIT_TESTING_FLAG

        self._setup_script_manager()
        appcontext_pushed.connect(self._install_id_generator)
        ServiceManager.__init__(self)
        PluginManager.__init__(self)
        self.default_view = ViewRegistry()
        self.js_api = dict()

        if config:
            self.config.from_object(config)

        # at this point we have loaded all external config files:
        # SQLALCHEMY_DATABASE_URI is definitively fixed (it cannot be defined in
        # database AFAICT), and LOGGING_FILE cannot be set in DB settings.
        self.setup_logging()

        configured = bool(self.config.get('SQLALCHEMY_DATABASE_URI'))
        self.config['CONFIGURED'] = configured

        if not self.testing:
            self.init_sentry()

        if not configured:
            # set fixed secret_key so that any unconfigured worker will use, so that
            # session can be used during setup even if multiple processes are
            # processing requests.
            self.config['SECRET_KEY'] = 'abilian_setup_key'

        # time to load config bits from database: 'settings'
        # First init required stuff: db to make queries, and settings service
        extensions.db.init_app(self)
        settings_service.init_app(self)

        if configured:
            with self.app_context():
                try:
                    settings = self.services['settings'].namespace(
                        'config').as_dict()
                except sa.exc.DatabaseError as exc:
                    # we may get here if DB is not initialized and "settings" table is
                    # missing. Command "initdb" must be run to initialize db, but first we
                    # must pass app init
                    if not self.testing:
                        # durint tests this message will show up on every test, since db is
                        # always recreated
                        logging.error(exc.message)
                    self.db.session.rollback()
                else:
                    self.config.update(settings)

        if not self.config.get('FAVICO_URL'):
            self.config['FAVICO_URL'] = self.config.get('LOGO_URL')

        languages = self.config.get('BABEL_ACCEPT_LANGUAGES')
        if languages is None:
            languages = abilian.i18n.VALID_LANGUAGES_CODE
        else:
            languages = tuple(lang for lang in languages
                              if lang in abilian.i18n.VALID_LANGUAGES_CODE)
        self.config['BABEL_ACCEPT_LANGUAGES'] = languages

        self._jinja_loaders = list()
        self.register_jinja_loaders(jinja2.PackageLoader('abilian.web',
                                                         'templates'))

        js_filters = (('closure_js',) if self.config.get('PRODUCTION', False)
                      else None)

        self._assets_bundles = {
            'css': {'options': dict(filters=('less', 'cssmin'),
                                    output='style-%(version)s.min.css',)},
            'js-top': {'options': dict(output='top-%(version)s.min.js',
                                       filters=js_filters,)},
            'js': {'options': dict(output='app-%(version)s.min.js',
                                   filters=js_filters)},
        }

        # bundles for JS translations
        for lang in languages:
            code = 'js-i18n-' + lang
            filename = 'lang-' + lang + '-%(version)s.min.js'
            self._assets_bundles[code] = {
                'options': dict(output=filename,
                                filters=js_filters),
            }

        for http_error_code in (403, 404, 500):
            self.install_default_handler(http_error_code)

        with self.app_context():
            self.init_extensions()
            self.register_plugins()
            self.add_access_controller('static',
                                       allow_access_for_roles(Anonymous),
                                       endpoint=True)
            # debugtoolbar: this is needed to have it when not authenticated on a
            # private site. We cannot do this in init_debug_toolbar, since auth
            # service is not yet installed
            self.add_access_controller('debugtoolbar',
                                       allow_access_for_roles(Anonymous),)
            self.add_access_controller('_debug_toolbar.static',
                                       allow_access_for_roles(Anonymous),
                                       endpoint=True)

        self.maybe_register_setup_wizard()
        self._finalize_assets_setup()
        # At this point all models should have been imported: time to configure
        # mappers. Normally Sqlalchemy does it when needed but mappers may be
        # configured inside sa.orm.class_mapper() which hides a misconfiguration: if
        # a mapper is misconfigured its exception is swallowed by
        # class_mapper(model) results in this laconic (and misleading) message:
        # "model is not mapped"
        sa.orm.configure_mappers()

        signals.components_registered.send(self)
        self.before_first_request(self._set_current_celery_app)
        self.before_first_request(lambda: signals.register_js_api.send(self))

        request_started.connect(self._setup_nav_and_breadcrumbs)

        # Initialize Abilian core services.
        # Must come after all entity classes have been declared.
        # Inherited from ServiceManager. Will need some configuration love later.
        if not self.config.get('TESTING', False):
            with self.app_context():
                self.start_services()

    def _setup_script_manager(self):
        manager = self.script_manager

        if manager is None or isinstance(manager, ScriptManager):
            return

        if isinstance(manager, (bytes, unicode)):
            manager = str(manager)
            if manager.startswith('.'):
                manager = self.import_name + manager

            manager_import_path = manager
            manager = import_string(manager, silent=True)
            if manager is None:
                # fallback on abilian-core's
                logger.warning(
                    '\n' + ('*' * 79) + '\n'
                    'Could not find command manager at %r, using a default one\n'
                    'Some commands might not be available\n' +
                    ('*' * 79) + '\n', manager_import_path)
                from abilian.core.commands import setup_abilian_commands
                manager = ScriptManager()
                setup_abilian_commands(manager)

            self.script_manager = manager

    def _install_id_generator(self, sender, **kwargs):
        g.id_generator = count(start=1)

    def _set_current_celery_app(self):
        """Listener for `before_first_request`.

        Set our celery app as current, so that task use the correct config.
        Without that tasks may use their default set app.
        """
        self.extensions['celery'].set_current()

    def _setup_nav_and_breadcrumbs(self, app=None):
        """Listener for `request_started` event.

        If you want to customize first items of breadcrumbs, override
        :meth:`init_breadcrumbs`
        """
        g.nav = {'active': None}  # active section
        g.breadcrumb = []
        self.init_breadcrumbs()

    def init_breadcrumbs(self):
        """Insert the first element in breadcrumbs.

        This happens during `request_started` event, which is triggered before any
        url_value_preprocessor and `before_request` handlers.
        """
        g.breadcrumb.append(BreadcrumbItem(icon=u'home',
                                           url=u'/' + request.script_root))

    def check_instance_folder(self, create=False):
        """Verify instance folder exists, is a directory, and has necessary permissions.

        :param:create: if `True`, creates directory hierarchy

        :raises: OSError with relevant errno
        """
        path = Path(self.instance_path)
        err = None
        eno = 0

        if not path.exists():
            if create:
                logger.info('Create instance folder: %s',
                            unicode(path).encode('utf-8'))
                path.mkdir(0o775, parents=True)
            else:
                err = 'Instance folder does not exists'
                eno = errno.ENOENT
        elif not path.is_dir():
            err = 'Instance folder is not a directory'
            eno = errno.ENOTDIR
        elif not os.access(str(path), os.R_OK | os.W_OK | os.X_OK):
            err = 'Require "rwx" access rights, please verify permissions'
            eno = errno.EPERM

        if err:
            raise OSError(eno, err, str(path))

        if not self.DATA_DIR.exists():
            self.DATA_DIR.mkdir(0o775, parents=True)

    def make_config(self, instance_relative=False):
        config = Flask.make_config(self, instance_relative)
        if not config.get('SESSION_COOKIE_NAME'):
            config['SESSION_COOKIE_NAME'] = self.name + '-session'

        # during testing DATA_DIR is not created by instance app, but we still need
        # this attribute to be set
        self.DATA_DIR = Path(self.instance_path, 'data')

        if self._ABILIAN_INIT_TESTING_FLAG:
            # testing: don't load any config file!
            return config

        if instance_relative:
            self.check_instance_folder(create=True)

        cfg_path = os.path.join(config.root_path, 'config.py')
        logger.info('Try to load config: "%s"', cfg_path)
        try:
            config.from_pyfile(cfg_path, silent=False)
        except IOError:
            return config

        config.from_envvar(self.CONFIG_ENVVAR, silent=True)

        if 'WTF_CSRF_ENABLED' not in config:
            config['WTF_CSRF_ENABLED'] = config.get('CSRF_ENABLED', True)

        return config

    def setup_logging(self):
        self.logger  # force flask to create application logger before logging
        # configuration; else, flask will overwrite our settings

        log_level = self.config.get("LOG_LEVEL")
        if log_level:
            self.logger.setLevel(log_level)

        logging_file = self.config.get('LOGGING_CONFIG_FILE')
        if logging_file:
            logging_file = os.path.abspath(os.path.join(self.instance_path,
                                                        logging_file))
        else:
            logging_file = resource_filename(__name__, 'default_logging.yml')

        if logging_file.endswith('.conf'):
            # old standard 'ini' file config
            logging.config.fileConfig(logging_file,
                                      disable_existing_loggers=False)
        elif logging_file.endswith('.yml'):
            # yml config file
            logging_cfg = yaml.load(open(logging_file, 'r'))
            logging_cfg.setdefault('version', 1)
            logging_cfg.setdefault('disable_existing_loggers', False)
            logging.config.dictConfig(logging_cfg)

    def init_debug_toolbar(self):
        if (not self.testing and self.config.get('DEBUG_TB_ENABLED') and
                'debugtoolbar' not in self.blueprints):
            try:
                from flask_debugtoolbar import DebugToolbarExtension, DebugToolbar
            except ImportError:
                logger.warning('DEBUG_TB_ENABLED is on but flask_debugtoolbar '
                               'is not installed.')
            else:
                try:
                    default_config = DebugToolbar.config
                    init_dbt = DebugToolbarExtension
                except AttributeError:
                    # debugtoolbar > 0.8.0
                    dbt = DebugToolbarExtension()
                    default_config = dbt._default_config(self)
                    init_dbt = dbt.init_app

                if 'DEBUG_TB_PANELS' not in self.config:
                    # add our panels to default ones
                    self.config['DEBUG_TB_PANELS'] = list(default_config[
                        'DEBUG_TB_PANELS'])
                    self.config['DEBUG_TB_PANELS'].append(
                        'abilian.services.indexing.debug_toolbar.IndexedTermsDebugPanel')
                init_dbt(self)
                for view_name in self.view_functions:
                    if view_name.startswith('debugtoolbar.'):
                        extensions.csrf.exempt(self.view_functions[view_name])

    def init_extensions(self):
        """Initialize flask extensions, helpers and services.
        """
        self.init_debug_toolbar()
        redis.Extension(self)
        extensions.mail.init_app(self)
        extensions.upstream_info.extension.init_app(self)
        actions.init_app(self)

        from abilian.core.jinjaext import DeferredJS
        DeferredJS(self)

        # auth_service installs a `before_request` handler (actually it's
        # flask-login). We want to authenticate user ASAP, so that sentry and logs
        # can report which user encountered any error happening later, in particular
        # in a before_request handler (like csrf validator)
        auth_service.init_app(self)

        # webassets
        self._setup_asset_extension()
        self._register_base_assets()

        # Babel (for i18n)
        abilian.i18n.babel.init_app(self)
        abilian.i18n.babel.add_translations('wtforms',
                                            translations_dir='locale',
                                            domain='wtforms')
        abilian.i18n.babel.add_translations('abilian')
        abilian.i18n.babel.localeselector(abilian.i18n.localeselector)
        abilian.i18n.babel.timezoneselector(abilian.i18n.timezoneselector)

        # Flask-Migrate
        Migrate(self, self.db)

        # CSRF by default
        if self.config.get('CSRF_ENABLED'):
            extensions.csrf.init_app(self)
            self.extensions['csrf'] = extensions.csrf
            extensions.abilian_csrf.init_app(self)

        self.register_blueprint(csrf.blueprint)

        # images blueprint
        from .web.views.images import blueprint as images_bp
        self.register_blueprint(images_bp)

        # Abilian Core services
        security_service.init_app(self)
        repository_service.init_app(self)
        session_repository_service.init_app(self)
        audit_service.init_app(self)
        index_service.init_app(self)
        activity_service.init_app(self)
        preferences_service.init_app(self)
        conversion_service.init_app(self)
        vocabularies_service.init_app(self)
        antivirus.init_app(self)

        from .web.preferences.user import UserPreferencesPanel
        preferences_service.register_panel(UserPreferencesPanel(), self)

        from .web.coreviews import users
        self.register_blueprint(users.bp)

        # Admin interface
        Admin().init_app(self)

        # Celery async service
        # this allows all shared tasks to use this celery app
        celery_app = self.extensions['celery'] = self.celery_app_cls()
        # force reading celery conf now - default celery app will
        # also update our config with default settings
        celery_app.conf  # noqa
        celery_app.set_default()

        # dev helper
        if self.debug:
            # during dev, one can go to /http_error/403 to see rendering of 403
            http_error_pages = Blueprint('http_error_pages', __name__)

            @http_error_pages.route('/<int:code>')
            def error_page(code):
                """ Helper for development to show 403, 404, 500..."""
                abort(code)

            self.register_blueprint(http_error_pages, url_prefix='/http_error')

    def register_plugins(self):
        """Load plugins listed in config variable 'PLUGINS'.
        """
        registered = set()
        for plugin_fqdn in chain(self.APP_PLUGINS, self.config['PLUGINS']):
            if plugin_fqdn not in registered:
                self.register_plugin(plugin_fqdn)
                registered.add(plugin_fqdn)

    def maybe_register_setup_wizard(self):
        if self.configured:
            return

        logger.info('Application is not configured, installing setup wizard')
        from abilian.web import setupwizard

        self.register_blueprint(setupwizard.setup, url_prefix='/setup')

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        """See :meth:`Flask.add_url_rule`.

        If `roles` parameter is present, it must be a
        :class:`abilian.service.security.models.Role` instance, or a list of
        Role instances.
        """
        roles = options.pop('roles', None)
        super(Application, self).add_url_rule(rule, endpoint, view_func,
                                              **options)

        if roles:
            self.add_access_controller(endpoint,
                                       allow_access_for_roles(roles),
                                       endpoint=True)

    def add_access_controller(self, name, func, endpoint=False):
        """Add an access controller.

        If `name` is None it is added at application level, else if is
        considered as a blueprint name. If `endpoint` is True then it is
        considered as an endpoint.
        """
        auth_state = self.extensions[auth_service.name]
        adder = auth_state.add_bp_access_controller

        if endpoint:
            adder = auth_state.add_endpoint_access_controller
            if not isinstance(name, string_types):
                raise ValueError('{} is not a valid endpoint name', repr(name))

        adder(name, func)

    def add_static_url(self, url_path, directory, endpoint=None, roles=None):
        """Add a new url rule for static files.

        :param endpoint: flask endpoint name for this url rule.
        :param url: subpath from application static url path. No heading or trailing
                    slash.
        :param directory: directory to serve content from.

        Example::

           app.add_static_url('myplugin',
                              '/path/to/myplugin/resources',
                              endpoint='myplugin_static')

        With default setup it will serve content from directory
        `/path/to/myplugin/resources` from url `http://.../static/myplugin`
        """
        url_path = self.static_url_path + '/' + url_path + '/<path:filename>'
        self.add_url_rule(url_path,
                          endpoint=endpoint,
                          view_func=partial(send_file_from_directory,
                                            directory=directory),
                          roles=roles)
        self.add_access_controller(endpoint,
                                   allow_access_for_roles(Anonymous),
                                   endpoint=True)

    #
    # Templating and context injection setup
    #
    def create_jinja_environment(self):
        env = Flask.create_jinja_environment(self)
        env.globals.update(app=current_app,
                           csrf=csrf,
                           get_locale=babel_get_locale,
                           local_dt=abilian.core.util.local_dt,
                           _n=abilian.i18n._n,
                           url_for=url_for,
                           user_photo_url=user_photo_url,
                           NO_VALUE=NO_VALUE,
                           NEVER_SET=NEVER_SET,)
        init_filters(env)
        return env

    @property
    def jinja_options(self):
        options = dict(Flask.jinja_options)

        extensions = options.setdefault('extensions', [])
        ext = 'abilian.core.jinjaext.DeferredJSExtension'
        if ext not in extensions:
            extensions.append(ext)

        if 'bytecode_cache' not in options:
            cache_dir = Path(self.instance_path, 'cache', 'jinja')
            if not cache_dir.exists():
                cache_dir.mkdir(0o775, parents=True)

            options['bytecode_cache'] = jinja2.FileSystemBytecodeCache(
                str(cache_dir), '%s.cache')

        if (self.config.get('DEBUG', False) and
                self.config.get('TEMPLATE_DEBUG', False)):
            options['undefined'] = jinja2.StrictUndefined
        return options

    def register_jinja_loaders(self, *loaders):
        """Register one or many `jinja2.Loader` instances for templates lookup.

        During application initialization plugins can register a loader so that
        their templates are available to jinja2 renderer.

        Order of registration matters: last registered is first looked up (after
        standard Flask lookup in app template folder). This allows a plugin to
        override templates provided by others, or by base application. The
        application can override any template from any plugins from its template
        folder (See `Flask.Application.template_folder`).

        :raise: `ValueError` if a template has already been rendered
        """
        if not hasattr(self, '_jinja_loaders'):
            raise ValueError(
                'Cannot register new jinja loaders after first template rendered')

        self._jinja_loaders.extend(loaders)

    @locked_cached_property
    def jinja_loader(self):
        """Search templates in custom app templates dir (default flask behaviour),
        fallback on abilian templates.
        """
        loaders = self._jinja_loaders
        del self._jinja_loaders
        loaders.append(Flask.jinja_loader.func(self))
        loaders.reverse()
        return jinja2.ChoiceLoader(loaders)

    # Error handling
    def handle_user_exception(self, e):
        # If session.transaction._parent is None, then exception has occured in
        # after_commit(): doing a rollback() raises an error and would hide actual
        # error
        session = db.session()
        if session.is_active and session.transaction._parent is not None:
            # inconditionally forget all DB changes, and ensure clean session during
            # exception handling
            session.rollback()
        else:
            self._remove_session_save_objects()

        return Flask.handle_user_exception(self, e)

    def handle_exception(self, e):
        session = db.session()
        if not session.is_active:
            # something happened in error handlers and session is not usable anymore.
            #
            self._remove_session_save_objects()

        return Flask.handle_exception(self, e)

    def _remove_session_save_objects(self):
        """
        Used during exception handling in case we need to remove() session: keep
        instances and merge them in the new session.
        """
        if self.testing:
            return
        # Before destroying the session, get all instances to be attached to the
        # new session. Without this, we get DetachedInstance errors, like when
        # tryin to get user's attribute in the error page...
        old_session = db.session()
        g_objs = []
        for key in iter(g):
            obj = getattr(g, key)
            if (isinstance(obj, db.Model) and
                    sa.orm.object_session(obj) in (None, old_session)):
                g_objs.append((key, obj, obj in old_session.dirty))

        db.session.remove()
        session = db.session()

        for key, obj, load in g_objs:
            # replace obj instance in bad session by new instance in fresh session
            setattr(g, key, session.merge(obj, load=load))

        # refresh `current_user`
        user = getattr(_request_ctx_stack.top, 'user', None)
        if user is not None and isinstance(user, db.Model):
            _request_ctx_stack.top.user = session.merge(user, load=load)

    def log_exception(self, exc_info):
        """Log exception only if sentry is not installed (this avoids getting error
        twice in sentry).
        """
        if 'sentry' not in self.extensions:
            super(Application, self).log_exception(exc_info)

    def init_sentry(self):
        """Install Sentry handler if config defines 'SENTRY_DSN'.
        """
        if self.config.get('SENTRY_DSN'):
            try:
                from abilian.core.sentry import Sentry
            except ImportError:
                logger.error(
                    'SENTRY_DSN is defined in config but package "raven" is not '
                    'installed.')
                return

            ext = Sentry(self, logging=True, level=logging.ERROR)
            ext.client.tags['app_name'] = self.name
            ext.client.tags['process_type'] = 'web'
            server_name = str(self.config.get('SERVER_NAME'))
            ext.client.tags['configured_server_name'] = server_name

    @property
    def db(self):
        return self.extensions['sqlalchemy'].db

    @property
    def redis(self):
        return self.extensions['redis'].client

    def create_db(self):
        from abilian.core.models.subjects import User

        db.create_all()
        if User.query.get(0) is None:
            root = User(id=0,
                        last_name=u'SYSTEM',
                        email=u'system@example.com',
                        can_login=False)
            db.session.add(root)
            db.session.commit()

    def _setup_asset_extension(self):
        assets = self.extensions['webassets'] = AssetsEnv(self)
        assets.debug = not self.config.get('PRODUCTION', False)
        assets.requirejs_config = {'waitSeconds': 90, 'shim': {}, 'paths': {},}

        assets_base_dir = Path(self.instance_path, 'webassets')
        assets_dir = assets_base_dir / 'compiled'
        assets_cache_dir = assets_base_dir / 'cache'
        for path in (assets_base_dir, assets_dir, assets_cache_dir):
            if not path.exists():
                path.mkdir()

        assets.directory = str(assets_dir)
        assets.cache = str(assets_cache_dir)
        manifest_file = assets_base_dir / 'manifest.json'
        assets.manifest = 'json:{}'.format(str(manifest_file))

        # set up load_path for application static dir. This is required since we are
        # setting Environment.load_path for other assets (like core_bundle below),
        # in this case Flask-Assets uses webasssets resolvers instead of Flask's one
        assets.append_path(self.static_folder, self.static_url_path)

        # filters options
        less_args = ['-ru']
        assets.config['less_extra_args'] = less_args
        assets.config['less_as_output'] = True
        if assets.debug:
            assets.config['less_source_map_file'] = 'style.map'

        # setup static url for our assets
        from abilian.web import assets as core_bundles
        core_bundles.init_app(self)

        # static minified are here
        assets.url = self.static_url_path + '/min'
        assets.append_path(str(assets_dir), assets.url)
        self.add_static_url('min',
                            str(assets_dir),
                            endpoint='webassets_static',
                            roles=Anonymous,)

    def _finalize_assets_setup(self):
        assets = self.extensions['webassets']
        assets_dir = Path(assets.directory)
        closure_base_args = [
            '--jscomp_warning',
            'internetExplorerChecks',
            '--source_map_format',
            'V3',
            '--create_source_map',
        ]

        for name, data in self._assets_bundles.items():
            bundles = data.get('bundles', [])
            options = data.get('options', {})
            filters = options.get('filters') or []
            options['filters'] = []
            for f in filters:
                if f == 'closure_js':
                    js_map_file = str(assets_dir / '{}.map'.format(name))
                    f = ClosureJS(extra_args=closure_base_args + [js_map_file])
                options['filters'].append(f)

            if not options['filters']:
                options['filters'] = None

            if bundles:
                assets.register(name, Bundle(*bundles, **options))

    def register_asset(self, type_, *assets):
        """Register webassets bundle to be served on all pages.

        :param type_: `"css"`, `"js-top"` or `"js""`.

        :param \*asset:
            a path to file, a :ref:`webassets.Bundle <webassets:bundles>` instance
            or a callable that returns a :ref:`webassets.Bundle <webassets:bundles>`
            instance.

        :raises KeyError: if `type_` is not supported.
        """
        supported = self._assets_bundles.keys()
        if type_ not in supported:
            raise KeyError("Invalid type: %s. Valid types: ", repr(type_),
                           ', '.join(sorted(supported)))

        for asset in assets:
            if not isinstance(asset, Bundle) and callable(asset):
                asset = asset()

            self._assets_bundles[type_].setdefault('bundles', []).append(asset)

    def register_i18n_js(self, *paths):
        """Register templates path translations files, like `select2/select2_locale_{lang}.js`.

        Only existing files are registered.
        """
        languages = self.config['BABEL_ACCEPT_LANGUAGES']
        assets = self.extensions['webassets']

        for path in paths:
            for lang in languages:
                filename = path.format(lang=lang)
                try:
                    assets.resolver.search_for_source(assets, filename)
                except IOError:
                    logger.debug('i18n JS not found, skipped: "%s"', filename)
                else:
                    self.register_asset('js-i18n-' + lang, filename)

    def _register_base_assets(self):
        """Register assets needed by Abilian.

        This is done in a separate method in order to allow applications to redefine it at will.
        """
        from abilian.web import assets as bundles

        self.register_asset('css', bundles.LESS)
        self.register_asset('js-top', bundles.TOP_JS)
        self.register_asset('js', bundles.JS)
        self.register_i18n_js(*bundles.JS_I18N)

    def install_default_handler(self, http_error_code):
        """Install a default error handler for `http_error_code`.

        The default error handler renders a template named error404.html for
        http_error_code 404.
        """
        logger.debug('Set Default HTTP error handler for status code %d',
                     http_error_code)
        handler = partial(self.handle_http_error, http_error_code)
        self.errorhandler(http_error_code)(handler)

    def handle_http_error(self, code, error):
        """Helper that renders `error{code}.html`.

        Convenient way to use it::

           from functools import partial
           handler = partial(app.handle_http_error, code)
           app.errorhandler(code)(handler)
        """
        # 5xx code: error on server side
        if (code / 100) == 5:
            # ensure rollback if needed, else error page may
            # have an error, too, resulting in raw 500 page :-(
            db.session.rollback()

        template = 'error{:d}.html'.format(code)
        return render_template(template, error=error), code


def create_app(config=None):
    return Application(config)



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from flask import current_app, request
from flask_login import user_logged_in
from raven.contrib.flask import Sentry as RavenExt


class Sentry(RavenExt):

    def init_app(self, app, *args, **kwargs):
        super(Sentry, self).init_app(app, *args, **kwargs)
        user_logged_in.connect(self._on_user_logged_in, sender=app)

    def _on_user_logged_in(self, app, user, *args, **kwargs):
        self.client.user_context(self.get_user_info(request))

    @property
    def raven_js_url(self):
        url = u'//cdn.ravenjs.com/{version}/{plugins}/raven.min.js'
        cfg = current_app.config
        return url.format(version=unicode(cfg['SENTRY_JS_VERSION']),
                          plugins=u','.join(cfg['SENTRY_JS_PLUGINS']))


__all__ = ['Sentry']



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from babel import Locale
from flask import _app_ctx_stack, _request_ctx_stack
from flask_babel import get_locale
from jinja2 import DictLoader

from abilian import i18n
from abilian.testing import BaseTestCase


class I18NTestCase(BaseTestCase):

    def test_set_locale(self):
        en = Locale('en')
        fr = Locale('fr')
        with self.app.test_request_context('/'):
            assert get_locale() == en

            with i18n.set_locale('fr') as new_locale:
                assert get_locale() == fr
                assert isinstance(new_locale, Locale)
                assert new_locale == fr
            assert get_locale() == en

            with i18n.set_locale(fr):
                assert get_locale() == fr
                with i18n.set_locale(en):
                    assert get_locale() == en
                assert get_locale() == fr
            assert get_locale() == en

        # no request context: no locale set
        app_ctx = _app_ctx_stack.top
        self._ctx.pop()
        self._ctx = None
        app_ctx.push()
        with i18n.set_locale(fr):
            assert get_locale() is None

    def test_ensure_request_context(self):
        en = Locale('en')
        fr = Locale('fr')

        # by default, a request context is present is test case setup.  first let's
        # show no new request context is set up and that set_locale works as
        # advertised
        app_ctx = _app_ctx_stack.top
        current_ctx = _request_ctx_stack.top
        assert current_ctx is not None
        assert get_locale() == en

        with i18n.ensure_request_context():
            assert current_ctx == _request_ctx_stack.top

            with i18n.set_locale(fr):
                assert get_locale() == fr

        self._ctx.pop()
        self._ctx = None
        assert _app_ctx_stack.top is None
        assert _request_ctx_stack.top is None
        assert get_locale() is None

        app_ctx.push()
        # ensure set_locale() doesn't set up a request context
        with i18n.set_locale(fr):
            assert _request_ctx_stack.top is None
            assert get_locale() is None

        with i18n.ensure_request_context():
            assert _request_ctx_stack.top is not None
            assert current_ctx != _request_ctx_stack.top
            assert get_locale() == en

            with i18n.set_locale(fr):
                assert get_locale() == fr

        assert _request_ctx_stack.top is None
        assert get_locale() is None

        # test chaining
        with i18n.ensure_request_context(), i18n.set_locale(fr):
            assert get_locale() == fr

    def test_get_template_i18n(self):
        template_path = '/myfile.txt'
        en = Locale('en')
        result = i18n.get_template_i18n(template_path, locale=en)
        self.assertIn('/myfile.en.txt', result)
        self.assertIn('/myfile.txt', result)

        en = Locale('en_US')
        result = i18n.get_template_i18n(template_path, locale=en)
        self.assertIn('/myfile.en_US.txt', result)
        self.assertIn('/myfile.txt', result)

        with i18n.set_locale('fr'):
            result = i18n.get_template_i18n(template_path, get_locale())
            self.assertIn('/myfile.fr.txt', result)
            self.assertIn('/myfile.txt', result)

    def test_render_template_i18n(self):
        loader = DictLoader({u'tmpl.txt': u'default ({{ locale }})',
                             u'tmpl.en.txt': u'en locale ({{ locale }})',
                             u'tmpl.fr.txt': u'fr locale ({{ locale }})',})
        app_loader = self.app.jinja_loader
        self.app.jinja_loader = loader
        render = i18n.render_template_i18n
        try:
            assert render('tmpl.txt', locale='fr') == u'fr locale (fr)'
            assert render('tmpl.txt', locale='en') == u'en locale (en)'
            assert render('tmpl.txt', locale='de') == u'default (de)'
        finally:
            self.app.jinja_loader = app_loader

    def test_default_country(self):
        assert 'DEFAULT_COUNTRY' in self.app.config
        assert self.app.config['DEFAULT_COUNTRY'] is None
        assert i18n.default_country() is None
        assert i18n.country_choices()[0][0] == u'AF'

        self.app.config['DEFAULT_COUNTRY'] = u'FR'
        assert i18n.default_country() == u'FR'
        assert i18n.country_choices()[0][0] == u'FR'
        assert i18n.country_choices(default_country_first=False)[0][0] == u'AF'



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from flask_login import AnonymousUserMixin

from abilian.services.indexing.schema import indexable_role
from abilian.services.security.models import Anonymous, Reader


def test_indexable_role():

    assert indexable_role(Anonymous) == u'role:anonymous'
    assert indexable_role(AnonymousUserMixin()) == u'role:anonymous'
    assert indexable_role(Reader) == u'role:reader'



# coding=utf-8
"""
Blueprint for views of dynamic images
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import colorsys
import hashlib

import pkg_resources
import sqlalchemy as sa
from flask import Blueprint, make_response, render_template, request
from pathlib import Path
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.models.blob import Blob
from abilian.core.models.subjects import User
from abilian.services.image import CROP, RESIZE_MODES, get_size
from abilian.web.util import url_for

from .files import BaseFileDownload

blueprint = Blueprint('images', __name__, url_prefix='/images')
route = blueprint.route

DEFAULT_AVATAR = Path(pkg_resources.resource_filename(
    'abilian.web', 'resources/img/avatar-default.png'))
DEFAULT_AVATAR_MD5 = hashlib.md5(DEFAULT_AVATAR.open('rb').read()).hexdigest()


class BaseImageView(BaseFileDownload):
    max_size = None

    def __init__(self, max_size=None, *args, **kwargs):
        super(BaseImageView, self).__init__(*args, **kwargs)
        # override class default value only if arg is specified in constructor. This
        # allows subclasses to easily override theses defaults.
        if max_size is not None:
            self.max_size = max_size

    def prepare_args(self, args, kwargs):
        args, kwargs = super(BaseImageView, self).prepare_args(args, kwargs)
        size = request.args.get('s', 0)
        try:
            size = int(size)
        except ValueError:
            raise BadRequest(
                'Invalid value for "s": {}. Not an integer.'.format(repr(size)))

        if self.max_size is not None:
            if size > self.max_size:
                raise BadRequest('Size too large: {:d} (max: {:d})'.format(
                    size, self.max_size))

        kwargs['size'] = size

        resize_mode = request.args.get('m', CROP)
        if resize_mode not in RESIZE_MODES:
            resize_mode = CROP

        kwargs['mode'] = resize_mode
        return args, kwargs

    def make_response(self, image, size, mode, *args, **kwargs):
        """
        :param image: image as bytes
        :param s: requested maximum width/height size
        """
        from abilian.services.image import resize, get_format

        try:
            fmt = get_format(image)
        except IOError:
            #  not a known image file
            raise NotFound()

        self.content_type = u'image/png' if fmt == 'PNG' else u'image/jpeg'
        ext = u'.' + unicode(fmt.lower())

        filename = kwargs.get('filename')
        if not filename:
            filename = u'image'
        if not filename.lower().endswith(ext):
            filename += ext
        self.filename = filename

        if size:
            image = resize(image, size, size, mode=mode)
            if mode == CROP:
                assert get_size(image) == (size, size)
        else:
            image = image.read()

        return make_response(image)

    def get_filename(self, *args, **kwargs):
        return self.filename

    def get_content_type(self, *args, **kwargs):
        return self.content_type


class StaticImageView(BaseImageView):
    """
    View for static assets not served by static directory.

    Useful for default avatars for example.
    """
    expire_vary_arg = 'md5'

    def __init__(self, image, *args, **kwargs):
        """
        :param image: path to image file
        """
        BaseImageView.__init__(self, *args, **kwargs)
        self.image_path = Path(image)
        if not self.image_path.exists():
            p = unicode(self.image_path)
            raise ValueError('Invalid image path: {}'.format(repr(p)))

    def prepare_args(self, args, kwargs):
        kwargs['image'] = self.image_path.open('rb')
        kwargs['filename'] = self.image_path.name
        return BaseImageView.prepare_args(self, args, kwargs)


class BlobView(BaseImageView):
    """
    Default :attr:`expire_vary_arg` to `"md5"`.
    :attr:`set_expire` is set to `False` by default.
    """
    expire_vary_arg = 'md5'
    id_arg = 'object_id'

    def __init__(self, id_arg=None, *args, **kwargs):
        BaseImageView.__init__(self, *args, **kwargs)
        if id_arg is not None:
            self.id_arg = id_arg

    def prepare_args(self, args, kwargs):
        args, kwargs = BaseImageView.prepare_args(self, args, kwargs)
        b_id = kwargs.get(self.id_arg)
        try:
            b_id = int(b_id)
        except ValueError:
            raise BadRequest('Invalid image id: {}'.format(repr(b_id)))

        blob = Blob.query.get(b_id)
        if not blob:
            raise NotFound()

        meta = blob.meta
        filename = meta.get('filename', meta.get('md5', unicode(blob.uuid)))
        kwargs['filename'] = filename
        kwargs['image'] = blob.file.open('rb')
        return args, kwargs


blob_image = BlobView.as_view(b'blob_image')
route("/files/<int:object_id>")(blob_image)


class UserMugshot(BaseImageView):
    expire_vary_arg = 'md5'

    def prepare_args(self, args, kwargs):
        args, kwargs = super(UserMugshot, self).prepare_args(args, kwargs)

        user_id = kwargs['user_id']
        user = User.query \
          .options(sa.orm.undefer(User.photo)) \
          .get(user_id)

        if user is None:
            raise NotFound()

        kwargs['user'] = user
        kwargs['image'] = user.photo
        return args, kwargs

    def make_response(self, user, image, size, *args, **kwargs):
        if image:
            #  user has set a photo
            return super(UserMugshot, self).make_response(image, size, *args,
                                                          **kwargs)

        # render svg avatar
        if user.last_name:
            letter = user.last_name[0]
        elif user.first_name:
            letter = user.first_name[0]
        else:
            letter = u"?"
        letter = letter.upper()
        # generate bg color, pastel: sat=65% in hsl color space
        id_hash = hash((user.name + user.email).encode('utf-8'))
        hue = id_hash % 10
        hue = (hue * 36) / 360.0  # 10 colors: 360 / 10
        color = colorsys.hsv_to_rgb(hue, 0.65, 1.0)
        color = [int(x * 255) for x in color]
        color = u'rgb({0[0]}, {0[1]}, {0[2]})'.format(color)
        svg = render_template('default/avatar.svg',
                              color=color,
                              letter=letter,
                              size=size)
        response = make_response(svg)
        self.content_type = u'image/svg+xml'
        self.filename = u'avatar-{}.svg'.format(id_hash)
        return response


user_photo = UserMugshot.as_view(b'user_photo', set_expire=True, max_size=500)
route("/users/<int:user_id>")(user_photo)
route('/users/default')(StaticImageView.as_view(b'user_default',
                                                set_expire=True,
                                                image=DEFAULT_AVATAR))


def user_url_args(user, size):
    endpoint = 'images.user_default'
    kwargs = {'s': size, 'md5': DEFAULT_AVATAR_MD5,}

    if not user.is_anonymous():
        endpoint = 'images.user_photo'
        kwargs['user_id'] = user.id
        content = (user.photo if user.photo else
                   (user.name + user.email).encode('utf-8'))
        kwargs['md5'] = hashlib.md5(content).hexdigest()

    return endpoint, kwargs


def user_photo_url(user, size):
    """Return url to use for this user.
    """
    endpoint, kwargs = user_url_args(user, size)
    return url_for(endpoint, **kwargs)



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sqlalchemy as sa

from abilian.core.entities import Entity
from abilian.testing import BaseTestCase


class IndexedContact(Entity):
    # default is 'test_service.IndexedContact'
    entity_type = 'abilian.services.indexing.IndexedContact'
    name = sa.Column(sa.UnicodeText)


class IndexingServiceTestCase(BaseTestCase):
    SERVICES = ('indexing',)

    def setUp(self):
        BaseTestCase.setUp(self)
        self.svc = self.app.services['indexing']

    def get_setup_config(self):
        cfg = BaseTestCase.get_setup_config(self)
        # cfg.SQLALCHEMY_ECHO = True
        return cfg

    def test_app_state(self):
        state = self.svc.app_state
        assert IndexedContact in state.indexed_classes
        assert IndexedContact.entity_type in self.svc.adapted
        assert IndexedContact.entity_type in state.indexed_fqcn

    def test_index_only_after_final_commit(self):
        contact = IndexedContact(name=u'John Doe')
        state = self.svc.app_state
        self.session.begin(nested=True)

        self.assertEquals(state.to_update, [])
        self.session.add(contact)

        # no commit: model is in wait queue
        self.session.flush()
        assert state.to_update == [('new', contact)]

        # commit but in a sub transaction: model still in wait queue
        self.session.commit()
        assert state.to_update == [('new', contact)]

        # 'final' commit: models sent for indexing update
        self.session.commit()
        assert state.to_update == []

    def test_clear(self):
        # just check no exception happens
        self.svc.clear()

        # check no double stop (would raise AssertionError from service base)
        self.svc.start()
        self.svc.stop()
        self.svc.clear()



# coding=utf-8
"""
Create all standard extensions.

"""

# Note: Because of issues with circular dependencies, Abilian-specific
# extensions are created later.

from __future__ import absolute_import, print_function, division

from abilian.core.logging import patch_logger
from sqlalchemy.engine import Engine

from flask import current_app
from . import upstream_info
from .login import login_manager
import flask_mail
import sqlalchemy as sa
from ..sqlalchemy import SQLAlchemy
from .csrf import wtf_csrf as csrf, abilian_csrf

__all__ = ['get_extension', 'db', 'mail', 'login_manager', 'csrf',
           'upstream_info']


# patch flask.ext.mail.Message.send to always set enveloppe_from default mail
# sender
# FIXME: we'ld rather subclass Message and update all imports
def _message_send(self, connection):
    """Send a single message instance.

    If TESTING is True the message will not actually be sent.

    :param message: a Message instance.
    """
    sender = current_app.config['MAIL_SENDER']
    if not self.extra_headers:
        self.extra_headers = {}
    self.extra_headers['Sender'] = sender
    connection.send(self, sender)


patch_logger.info(flask_mail.Message.send)
flask_mail.Message.send = _message_send

mail = flask_mail.Mail()

db = SQLAlchemy()


@sa.event.listens_for(db.metadata, 'before_create')
@sa.event.listens_for(db.metadata, 'before_drop')
def _filter_metadata_for_connection(target, connection, **kw):
    """Listener to control what indexes get created.

    Useful for skipping postgres-specific indexes on a sqlite for example.

    It's looking for info entry `engines` on an index
    (`Index(info=dict(engines=['postgresql']))`), an iterable of engine names.
    """
    engine = connection.engine.name
    default_engines = (engine,)
    tables = target if isinstance(target, sa.Table) else kw.get('tables', [])
    for table in tables:
        indexes = list(table.indexes)
        for idx in indexes:
            if engine not in idx.info.get('engines', default_engines):
                table.indexes.remove(idx)

# csrf


def get_extension(name):
    """Get the named extension from the current app, returning None if not found.
    """
    return current_app.extensions.get(name)


def _install_get_display_value(cls):

    _MARK = object()

    def display_value(self, field_name, value=_MARK):
        """ Return display value for fields having 'choices' mapping (stored value
        -> human readable value). For other fields it will simply return field
        value.

        `display_value` should be used instead of directly getting field value.

        If `value` is provided it is "tranlated" to a human-readable value. This is
        useful for obtaining a human readable label from a raw value
        """
        val = getattr(self, field_name) if value is _MARK else value

        mapper = sa.orm.object_mapper(self)
        try:
            field = getattr(mapper.c, field_name)
        except AttributeError:
            pass
        else:
            if 'choices' in field.info:
                get = lambda v: field.info['choices'].get(v, v)
                if isinstance(val, list):
                    val = [get(v) for v in val]
                else:
                    val = get(val)

        return val

    if not hasattr(cls, 'display_value'):
        cls.display_value = display_value


sa.event.listen(db.Model, 'class_instrument', _install_get_display_value)


#
# Make Sqlite a bit more well-behaved.
#
@sa.event.listens_for(Engine, "connect")
def _set_sqlite_pragma(dbapi_connection, connection_record):
    from sqlite3 import Connection as SQLite3Connection
    if isinstance(dbapi_connection, SQLite3Connection):  # pragma: no cover
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from flask import g
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from whoosh.fields import STORED

from .base import AUDITABLE, EDITABLE, SEARCHABLE, SYSTEM
from .subjects import User


class OwnedMixin(object):
    __indexation_args__ = {
        'index_to': (('creator', ('creator',)),
                     ('creator_name', (('creator_name', STORED),)),
                     ('owner', ('owner',)),
                     ('owner_name', (('owner_name', STORED),)),)
    }

    def __init__(self, *args, **kwargs):
        if hasattr(g, 'user'):
            if not self.creator and not g.user.is_anonymous():
                self.creator = g.user
            if not self.owner and not g.user.is_anonymous():
                self.owner = g.user

    @declared_attr
    def creator_id(cls):
        return Column(ForeignKey(User.id), info=SYSTEM)

    @declared_attr
    def creator(cls):
        pj = "User.id == %s.creator_id" % cls.__name__
        return relationship(User,
                            primaryjoin=pj,
                            lazy='joined',
                            uselist=False,
                            info=SYSTEM | SEARCHABLE)

    @property
    def creator_name(self):
        return unicode(self.creator) if self.creator else u''

    @declared_attr
    def owner_id(cls):
        return Column(ForeignKey(User.id), info=EDITABLE | AUDITABLE)

    @declared_attr
    def owner(cls):
        pj = "User.id == %s.owner_id" % cls.__name__
        return relationship(User,
                            primaryjoin=pj,
                            lazy='joined',
                            uselist=False,
                            info=EDITABLE | AUDITABLE | SEARCHABLE)

    @property
    def owner_name(self):
        return unicode(self.owner) if self.owner else u''



# coding=utf-8
"""
Class based views
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

import sqlalchemy as sa
from flask import (current_app, flash, g, redirect, render_template, request,
                   url_for)
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.entities import ValidationError
from abilian.core.signals import activity
from abilian.i18n import _, _l
from abilian.services.security import CREATE, DELETE, READ, WRITE

from .. import csrf, forms, nav
from ..action import ButtonAction, Endpoint, actions
from .base import JSONView, View

logger = logging.getLogger(__name__)


class BaseObjectView(View):
    """
    Base class common to all database objects views.
    """
    #: form title
    title = None

    #: Model class
    Model = None

    #: primary key name to look for in url arguments
    pk = 'object_id'

    #: object instance for this view
    obj = None

    #: object id
    object_id = None

    #: template to render
    template = None

    #: default templates inherit from "base_template". This allows to use generic
    #: templates with a custom base
    base_template = "base.html"

    def __init__(self,
                 Model=None,
                 pk=None,
                 base_template=None,
                 *args,
                 **kwargs):
        View.__init__(self, *args, **kwargs)
        cls = self.__class__
        self.pk = pk if pk is not None else cls.pk
        self.Model = Model if Model is not None else cls.Model
        self.base_template = (base_template if base_template is not None else
                              cls.base_template)

    def prepare_args(self, args, kwargs):
        args, kwargs = self.init_object(args, kwargs)

        if self.obj is None:
            raise NotFound()

        return args, kwargs

    def breadcrumb(self):
        """
        Return :class:`..nav.BreadcrumbItem` instance for this object.

        This method may return a list of BreadcrumbItem instances. Return
        `None` if nothing.
        """
        return None

    def init_object(self, args, kwargs):
        """
        This method is reponsible for setting :attr:`obj`. It is called during
        :meth:`prepare_args`.
        """
        self.object_id = kwargs.pop(self.pk, None)
        if self.object_id is not None:
            self.obj = self.Model.query.get(self.object_id)
            actions.context['object'] = self.obj

        return args, kwargs

    def get(self, *args, **kwargs):
        bc = self.breadcrumb()
        if bc is not None:
            bc = [bc] if isinstance(bc, nav.BreadcrumbItem) else list(bc)
            assert all(isinstance(b, nav.BreadcrumbItem) for b in bc)
            g.breadcrumb.extend(bc)

        kwargs = {'base_template': self.base_template}
        kwargs.update(self.template_kwargs)
        # forbid override "view"
        kwargs['view'] = self
        return render_template(self.template, **kwargs)

    @property
    def template_kwargs(self):
        """
        Template render arguments. You can override `base_template` for
        instance. Only `view` cannot be overriden.
        """
        return {}


class ObjectView(BaseObjectView):
    """View objects.
    """

    #: html template
    template = 'default/object_view.html'

    #: View form class. Form object used to show objects fields
    Form = None

    #: required permission. Must be an instance of
    #: :class:`abilian.services.security.Permission`
    permission = READ

    #: form instance for this view
    form = None

    def __init__(self,
                 Model=None,
                 pk=None,
                 Form=None,
                 template=None,
                 *args,
                 **kwargs):
        super(ObjectView, self).__init__(Model, pk, *args, **kwargs)
        cls = self.__class__
        self.Form = Form if Form is not None else cls.Form
        self.template = template if template is not None else cls.template

    def prepare_args(self, args, kwargs):
        """
        :attr:`form` is initialized here. See also :meth:`View.prepare_args`.
        """
        args, kwargs = super(ObjectView, self).prepare_args(args, kwargs)
        self.form = self.Form(**self.get_form_kwargs())
        return args, kwargs

    def get_form_kwargs(self):
        kw = dict(obj=self.obj)
        if issubclass(self.Form, forms.Form) and self.permission:
            kw['permission'] = self.permission
        return kw

    def index_url(self):
        return url_for('.index')

    def redirect_to_index(self):
        return redirect(self.index_url())

    @property
    def template_kwargs(self):
        """Provides :attr:`form` to templates
        """
        kw = super(ObjectView, self).template_kwargs
        kw['form'] = self.form
        return kw


CANCEL_BUTTON = ButtonAction(
    'form',
    'cancel',
    title=_l(u'Cancel'),
    btn_class='default cancel'  # .cancel: if jquery.validate is used it will
)  # properly skip validation

EDIT_BUTTON = ButtonAction(
    'form', 'edit', btn_class='primary',
    title=_l(u'Save'))


class ObjectEdit(ObjectView):
    """Edit objects.
    """
    template = 'default/object_edit.html'
    decorators = (csrf.support_graceful_failure,)
    permission = WRITE

    #: :class:ButtonAction instance to show on form
    _buttons = ()

    #: submitted form data
    data = None

    #: action name from form data
    action = None

    #: button clicked, corresponding to :attr:`action`.
    button = None

    #: verb used to describe activity
    activity_verb = 'update'

    #: UI flash message
    _message_success = _l(u"Entity successfully edited")

    view_endpoint = None

    def __init__(self,
                 Model=None,
                 pk=None,
                 Form=None,
                 template=None,
                 view_endpoint=None,
                 message_success=None,
                 *args,
                 **kwargs):
        ObjectView.__init__(
            self, Model,
            pk, Form, template=template,
            *args, **kwargs)
        if view_endpoint is not None:
            self.view_endpoint = view_endpoint

        if not self.view_endpoint:
            self.view_endpoint = '.{}_view'.format(self.Model.__name__)

        if message_success:
            self._message_success = message_success

    def post(self, *args, **kwargs):
        # conservative: no action submitted -> cancel
        action = self.data.get('__action', u'cancel')
        if action == u'cancel':
            return self.cancel()

        return self.handle_action(action)

    def put(self):
        return self.post()

    def prepare_args(self, args, kwargs):
        args, kwargs = super(ObjectEdit, self).prepare_args(args, kwargs)
        self._buttons = self.get_form_buttons(*args, **kwargs)
        self.data = request.form
        return args, kwargs

    def get_form_buttons(self, *args, **kwargs):
        return [EDIT_BUTTON, CANCEL_BUTTON]

    @property
    def buttons(self):
        return (button for button in self._buttons
                if button.available(actions.context))

    def view_url(self):
        kw = {self.pk: self.obj.id}
        return url_for(self.view_endpoint, **kw)

    def redirect_to_view(self):
        if self.button:
            url = self.button.url(actions.context)
            if url:
                return redirect(url)
        return redirect(self.view_url())

    def message_success(self):
        return unicode(self._message_success)

    # actions
    def handle_action(self, action):
        for button in self._buttons:
            if action == button.name:
                if not button.available(dict(view=self)):
                    raise ValueError('Action "{}" not available'
                                     ''.format(action.encode('utf-8')))
                break
        else:
            raise ValueError('Unknown action: "{}"'.format(action.encode(
                'utf-8')))

        self.action = action
        self.button = button
        return getattr(self, action)()

    def cancel(self):
        return self.redirect_to_view()

    def edit(self):
        if self.validate():
            return self.form_valid()
        else:
            if request.csrf_failed:
                errors = self.form.errors
                csrf_failed = errors.pop('csrf_token', False)
                if csrf_failed and not errors:
                    # failed only because of invalid/expired csrf, no error on form
                    return self.form_csrf_invalid()

            resp = self.form_invalid()
            if resp:
                return resp

            flash(_(u"Please fix the error(s) below"), "error")

        # if we end here then something wrong has happened: show form with error
        # messages
        return self.get()

    def before_populate_obj(self):
        """
        This method is called after form has been validated and before calling
        `form.populate_obj()`. Sometimes one may want to remove a field from
        the form because it's non-sense to store it on edited object, and use it in
        a specific manner, for example::

            image = form.image
            del form.image
            store_image(image)
        """
        pass

    def after_populate_obj(self):
        """
        Called after `self.obj` values have been updated, and `self.obj`
        attached to an ORM session.
        """
        pass

    def handle_commit_exception(self, exc):
        """
        hook point to handle exception that may happen during commit.

        It is the responsability of this method to perform a rollback if it is
        required for handling `exc`. If the method does not handle `exc` if should
        do nothing and return None.

        :returns: * a valid :class:`Response` if exception is handled.
                  * `None` if exception is not handled. Default handling happens.
        """
        return None

    def commit_success(self):
        """
        Called after object has been successfully saved to database
        """

    def validate(self):
        return self.form.validate()

    def form_valid(self):
        """Save object.

        Called when form is validated.
        """
        session = current_app.db.session()

        with session.no_autoflush:
            self.before_populate_obj()
            self.form.populate_obj(self.obj)
            session.add(self.obj)
            self.after_populate_obj()

        try:
            session.flush()
            self.send_activity()
            session.commit()
        except ValidationError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            flash(e.message, "error")
            return self.get()
        except sa.exc.IntegrityError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            logger.error(e)
            flash(_(u"An entity with this name already exists in the system."),
                  "error")
            return self.get()
        else:
            self.commit_success()
            flash(self.message_success(), "success")
            return self.redirect_to_view()

    def form_invalid(self):
        """
        When a form doesn't validate this method is called.

        It may return a :class:`Flask.Response` instance, to handle specific
        errors in custom screens.

        Else the edit form screen is returned with error(s) highlighted.

        This method is useful for detecting edition conflict using hidden fields
        and show a specific screen to help resolve the conflict.
        """
        return None

    def form_csrf_invalid(self):
        """
        Called when a form doesn't validate *only* because of csrf token expiration.

        This works only if form is an instance of :class:`flask_wtf.form.SecureForm`.
        Else default CSRF protection (before request) will take place.

        It must return a valid :class:`Flask.Response` instance. By default it
        returns to edit form screen with an informative message.
        """
        current_app.extensions['csrf-handler'].flash_csrf_failed_message()
        return self.get()

    def send_activity(self):
        activity.send(self,
                      actor=g.user,
                      verb=self.activity_verb,
                      object=self.obj,
                      target=self.activity_target)

    @property
    def activity_target(self):
        """
        Return `target` to use when creating activity.
        """
        return None


CREATE_BUTTON = ButtonAction(
    'form', 'create', btn_class='primary',
    title=_l(u'Create'))
CHAIN_CREATE_BUTTON = ButtonAction(
    'form',
    'chain_create',
    btn_class='primary',
    title=_l(u'Create and add new'),
    endpoint=lambda ctx: Endpoint(request.endpoint, **request.view_args),
    condition=lambda ctx: getattr(ctx['view'], 'chain_create_allowed', False))


class ObjectCreate(ObjectEdit):
    """Create a new object.
    """
    permission = CREATE
    activity_verb = 'post'
    _message_success = _l(u"Entity successfully added")

    #: set to `True` to show 'Save and add new' button
    chain_create_allowed = False

    def __init__(self, *args, **kwargs):
        chain_create_allowed = kwargs.pop('chain_create_allowed', None)
        if chain_create_allowed is not None:
            self.chain_create_allowed = bool(chain_create_allowed)

        ObjectEdit.__init__(self, *args, **kwargs)

    def prepare_args(self, args, kwargs):
        # we must ensure that no flush() occurs and that obj is not registered in
        # session (to prevent accidental insert of an incomplete object)
        session = current_app.db.session()
        with session.no_autoflush:
            args, kwargs = super(ObjectCreate, self).prepare_args(args, kwargs)

        try:
            session.expunge(self.obj)
        except sa.exc.InvalidRequestError:
            # obj is not in session
            pass

        return args, kwargs

    def init_object(self, args, kwargs):
        self.obj = self.Model()
        return args, kwargs

    def get_form_kwargs(self):
        kw = super(ObjectCreate, self).get_form_kwargs()
        if request.method == 'GET':
            # when GET allow form prefill instead of empty/current object data
            # FIXME: filter allowed parameters on given a field flags (could be
            # 'allow_from_get'?)
            kw['formdata'] = request.args

        return kw

    def get_form_buttons(self, *args, **kwargs):
        return [CREATE_BUTTON, CHAIN_CREATE_BUTTON, CANCEL_BUTTON]

    def breadcrumb(self):
        return nav.BreadcrumbItem(label=CREATE_BUTTON.title)

    # actions
    def create(self):
        return self.edit()

    chain_create = create

    def cancel(self):
        return self.redirect_to_index()


DELETE_BUTTON = ButtonAction('form', 'delete', title=_l(u'Delete'))


class ObjectDelete(ObjectEdit):
    """Delete object. Supports DELETE verb.
    """
    methods = ['POST']
    permission = DELETE
    activity_verb = 'delete'
    _message_success = _l(u"Entity deleted")

    init_object = BaseObjectView.init_object

    def get_form_buttons(self, *args, **kwargs):
        return [DELETE_BUTTON, CANCEL_BUTTON]

    def delete(self):
        session = current_app.db.session()
        session.delete(self.obj)
        activity.send(self,
                      actor=g.user,
                      verb="delete",
                      object=self.obj,
                      target=self.activity_target)
        try:
            session.commit()
        except sa.exc.IntegrityError as e:
            rv = self.handle_commit_exception(e)
            if rv is not None:
                return rv
            session.rollback()
            logger.error(e)
            flash(_("This entity is referenced by another object and cannot be deleted."),
                  "error")
            return self.redirect_to_view()
        else:
            flash(self.message_success(), 'success')
            # FIXME: for DELETE verb response in case of success should be 200, 202
            # (accepted) or 204 (no content)
            return self.redirect_to_index()


class JSONBaseSearch(JSONView):
    Model = None
    minimum_input_length = 2

    def __init__(self, *args, **kwargs):
        Model = kwargs.pop('Model', self.Model)
        minimum_input_length = kwargs.pop('minimum_input_length',
                                          self.minimum_input_length)
        super(JSONBaseSearch, self).__init__(*args, **kwargs)
        self.Model = Model
        self.minimum_input_length = minimum_input_length

    def prepare_args(self, args, kwargs):
        args, kwargs = JSONView.prepare_args(self, args, kwargs)
        kwargs['q'] = kwargs.get("q", u'').replace(u"%", u" ").lower()
        return args, kwargs

    def data(self, q, *args, **kwargs):
        if self.minimum_input_length and len(q) < self.minimum_input_length:
            raise BadRequest('Minimum query length is {:d}'.format(
                self.minimum_input_length),)

        results = []
        for obj in self.get_results(q, **kwargs):
            results.append(self.get_item(obj))

        return dict(results=results)

    def get_results(self, q, *args, **kwargs):
        raise NotImplementedError

    def get_item(self, obj):
        """
        Return a result item

        :param obj: Instance object
        :returns: a dictionnary with at least `id` and `text` values
        """
        raise NotImplementedError


class JSONModelSearch(JSONBaseSearch):
    """
    Base class for json sqlalchemy model search, as used by select2 widgets for
    example
    """

    def get_results(self, q, *args, **kwargs):
        query = self.Model.query

        query = self.options(query)
        query = self.filter(query, q, **kwargs)
        query = self.order_by(query)
        if not q and not self.minimum_input_length:
            query = query.limit(50)
        return query.all()

    def options(self, query):
        return query.options(sa.orm.noload('*'))

    def filter(self, query, q, **kwargs):
        if not q:
            return query
        return query.filter(sa.func.lower(self.Model.name).like(q + "%"))

    def order_by(self, query):
        return query.order_by(self.Model.name)

    def get_label(self, obj):
        return obj.name

    def get_item(self, obj):
        """
        Return a result item.

        :param obj: Instance object
        :returns: a dictionnary with at least `id` and `text` values
        """
        return dict(id=obj.id, text=self.get_label(obj), name=obj.name)


class JSONWhooshSearch(JSONBaseSearch):
    """
    Base class for JSON Whoosh search, as used by select2 widgets for example
    """

    def get_results(self, q, *args, **kwargs):
        svc = current_app.services['indexing']
        search_kwargs = {'limit': 30, 'Models': (self.Model,)}
        results = svc.search(q, **search_kwargs)
        return results

    def get_item(self, hit):
        """Return a result item.

        :param hit: Hit object from Whoosh
        :returns: a dictionnary with at least `id` and `text` values
        """
        return dict(id=hit['id'], text=hit['name'], name=hit['name'])



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sqlalchemy as sa

from abilian.testing import BaseTestCase
from abilian.web import url_for

from .models import BaseVocabulary, Vocabulary
from .service import vocabularies


class TestVocabularies(BaseTestCase):

    DefaultVoc = Vocabulary(b'defaultstates', label=u'States')

    def test_vocabulary_creator(self):
        PriorityVoc = Vocabulary(b'priorities', label=u'Priorities')
        assert PriorityVoc.__name__ == 'VocabularyPriorities'
        assert PriorityVoc.__tablename__ == 'vocabulary_priorities'
        assert PriorityVoc.Meta.name == u'priorities'
        assert PriorityVoc.Meta.label == u'Priorities'
        assert PriorityVoc.Meta.group is None
        assert issubclass(PriorityVoc, BaseVocabulary)

        StateVoc = self.DefaultVoc
        DocCatVoc = Vocabulary(b'categories',
                               group='documents',
                               label=u'Categories')

        # test registered vocabularies
        assert vocabularies.vocabularies == {PriorityVoc, StateVoc, DocCatVoc}
        assert (
            vocabularies.grouped_vocabularies == {None: [StateVoc, PriorityVoc],
                                                  'documents': [DocCatVoc]})
        assert vocabularies.get_vocabulary('priorities') is PriorityVoc
        assert vocabularies.get_vocabulary('priorities', 'nogroup') is None
        assert vocabularies.get_vocabulary('categories',
                                           'documents') is DocCatVoc

        self.app.db.create_all()

        IMMEDIATE = PriorityVoc(label=u'Immediate', position=0)
        NORMAL = PriorityVoc(label=u'Normal', position=3, default=True)
        URGENT = PriorityVoc(label=u'Urgent', position=1)
        HIGH = PriorityVoc(label=u'High', position=2)
        items = (IMMEDIATE, NORMAL, URGENT, HIGH)
        map(self.session.add, items)
        self.session.flush()

        # test position=4 set automatically; Label stripped
        low_item = PriorityVoc(label=u' Low  ')
        self.session.add(low_item)
        self.session.flush()
        assert low_item.position == 4
        assert low_item.label == u'Low'

        # test strip label on update
        IMMEDIATE.label = u'  Immediate  '
        self.session.flush()
        assert IMMEDIATE.label == u'Immediate'

        # test default ordering
        assert ([i.label for i in PriorityVoc.query.active().all()] ==
                [u'Immediate', u'Urgent', u'High', u'Normal', u'Low'])

        # no default ordering when using .values(): explicit ordering required
        query = PriorityVoc.query.active().order_by(PriorityVoc.position.asc())
        assert ([i.label for i in query.values(PriorityVoc.label)] ==
                [u'Immediate', u'Urgent', u'High', u'Normal', u'Low'])

        # test db-side constraint for non-empty labels
        try:
            with self.session.begin_nested():
                v = PriorityVoc(label=u'   ', position=6)
                self.session.add(v)
                self.session.flush()
        except sa.exc.IntegrityError:
            pass
        else:
            self.fail("Could insert an item with empty label")

        # test unique labels constraint
        try:
            with self.session.begin_nested():
                v = PriorityVoc(label=u'Immediate')
                self.session.add(v)
                self.session.flush()
        except sa.exc.IntegrityError:
            pass
        else:
            self.fail("Could insert duplicate label")

        # test unique position constraint
        try:
            with self.session.begin_nested():
                v = PriorityVoc(label=u'New one', position=1)
                self.session.add(v)
                self.session.flush()
        except sa.exc.IntegrityError:
            pass
        else:
            self.fail("Could insert duplicate position")

        # test by_position without results
        item = PriorityVoc.query.by_position(42)
        assert item is None

        # test by_position() and active()
        item = PriorityVoc.query.by_position(URGENT.position)
        assert item is URGENT
        item.active = False
        assert ([i.label for i in PriorityVoc.query.active().all()] ==
                [u'Immediate', u'High', u'Normal', u'Low'])
        assert PriorityVoc.query.active().by_position(URGENT.position) is None

        # test by_label()
        item = PriorityVoc.query.by_label(NORMAL.label)
        assert item is NORMAL

    def test_admin_panel_reorder(self):
        Voc = self.DefaultVoc
        session = self.session
        items = [
            Voc(label=u'First',
                position=0), Voc(label=u'Second',
                                 position=2), Voc(label=u'Third',
                                                  position=3)
        ]

        for i in items:
            session.add(i)
        session.commit()

        first, second, third = items
        url = url_for('admin.vocabularies')
        base_data = {'Model': Voc.Meta.name}
        data = {'down': first.id}
        data.update(base_data)
        r = self.client.post(url, data=data)
        assert r.status_code == 302
        assert r.headers['Location'] == u'http://localhost/admin/vocabularies'
        assert Voc.query.order_by(Voc.position).all() == [second, first, third]

        data = {'up': first.id, 'return_to': 'group'}
        data.update(base_data)
        r = self.client.post(url, data=data)
        assert r.status_code == 302
        assert r.headers[
            'Location'] == u'http://localhost/admin/vocabularies/_/'
        assert Voc.query.order_by(Voc.position).all() == [first, second, third]

        data = {'up': first.id, 'return_to': 'model'}
        data.update(base_data)
        r = self.client.post(url, data=data)
        assert r.status_code == 302
        assert (r.headers['Location'] ==
                u'http://localhost/admin/vocabularies/_/defaultstates/')
        assert Voc.query.order_by(Voc.position).all() == [first, second, third]

        data = {'down': third.id}
        data.update(base_data)
        r = self.client.post(url, data=data)
        assert r.status_code == 302
        assert r.headers['Location'] == u'http://localhost/admin/vocabularies'
        assert Voc.query.order_by(Voc.position).all() == [first, second, third]

        data = {'up': third.id}
        data.update(base_data)
        r = self.client.post(url, data=data)
        assert r.status_code == 302
        assert r.headers['Location'] == u'http://localhost/admin/vocabularies'
        assert Voc.query.order_by(Voc.position).all() == [first, third, second]



# coding=utf-8
"""
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from wtforms.fields import StringField

from abilian.core.models.attachment import Attachment
from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.fields import FileField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required


class AttachmentForm(Form):

    blob = FileField(
        _l(u'file'),
        validators=[required()],
        filters=(strip,),
        multiple=False)

    description = StringField(_l(u'description (optional)'), filters=(strip,),)

    class Meta:
        model = Attachment
        include_primary_keys = True
        assign_required = False  # for 'id': allow None, for new records



# coding=utf-8
"""
Jinja2 extensions
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from functools import partial

import lxml.html
from flask import current_app
from flask.globals import _lookup_req_object, _request_ctx_stack
from flask.signals import got_request_exception, request_started
from jinja2 import nodes
from jinja2.ext import Extension
from werkzeug.local import LocalProxy

deferred_js = LocalProxy(partial(_lookup_req_object, 'deferred_js'))


class DeferredJS(object):
    """Flask extentions for use with DeferredJSExtension for jinja
    """
    name = 'deferred_js'

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if self.name in app.extensions:
            return

        app.extensions[self.name] = self
        request_started.connect(self.reset_request, app)
        got_request_exception.connect(self.reset_request, app)

    def reset_request(self, sender, **extra):
        self.reset_deferred()

    def reset_deferred(self):
        _request_ctx_stack.top.deferred_js = []


class DeferredJSExtension(Extension):
    """Put JS fragment at the end of the document in a script tag.

    The JS fragment can contains <script> tag so that your favorite editor
    keeps doing proper indentation, syntax highlighting...
    """
    tags = {'deferJS', 'deferredJS'}

    def parse(self, parser):
        token = next(parser.stream)
        tag = token.value
        lineno = token.lineno

        # now we parse body of the block
        body = parser.parse_statements(
            ['name:enddeferJS', 'name:enddeferredJS'],
            drop_needle=True)

        method = 'defer_nodes' if tag == 'deferJS' else 'collect_deferred'
        return nodes.CallBlock(
            self.call_method(method, []), [], [], body).set_lineno(lineno)

    def defer_nodes(self, caller):
        body = u'<div>{}</div>'.format(caller().strip())

        # remove 'script' tag in immediate children, if any
        fragment = lxml.html.fragment_fromstring(body)
        for child in fragment:
            if child.tag == 'script':
                child.drop_tag()  # side effect on fragment.text or previous_child.tail!

        body = []
        if fragment.text:
            body.append(fragment.text)

        for child in fragment:
            body.append(lxml.html.tostring(child))
            body.append(child.tail)
        body = u''.join(body)

        deferred_js.append(body)
        return u''

    def collect_deferred(self, caller):
        result = u'\n'.join(u'(function(){{\n{}\n}})();'.format(js)
                            for js in deferred_js)
        current_app.extensions[DeferredJS.name].reset_deferred()
        return result



"""
Models for user preferences.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import backref, relation

from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.core.sqlalchemy import JSON


class UserPreference(db.Model):
    """An atom of user preference."""
    __tablename__ = 'user_preference'
    __table_args__ = (UniqueConstraint('user_id', 'key'),)

    #: Unique id for this preference.
    id = Column(Integer, primary_key=True, autoincrement=True)

    #: The user who set this preference.
    user = relation(User,
                    backref=backref('preferences',
                                    cascade="all, delete, delete-orphan"))
    user_id = Column(ForeignKey(User.id))

    #: The key
    key = Column(String, nullable=False)

    #: The value
    value = Column(JSON, nullable=False)
